# Copyright (C) 2007-2012 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""Interfaces for the request database.

The request database handles events that must be approved by the list
moderators, such as subscription requests and held messages.
"""

from __future__ import absolute_import, unicode_literals

__metaclass__ = type
__all__ = [
    'IListRequests',
    'RequestType',
    ]


from flufl.enum import Enum
from zope.interface import Interface, Attribute



class RequestType(Enum):
    held_message = 1
    subscription = 2
    unsubscription = 3



class IListRequests(Interface):
    """Held requests for a specific mailing list."""

    mailing_list = Attribute(
        """The IMailingList for these requests.""")

    count = Attribute(
        """The total number of requests held for the mailing list.""")

    def count_of(request_type):
        """The total number of requests held of the given request type.

        :param request_type: A `RequestType` enum value.
        :return: An integer.
        """

    def hold_request(request_type, key, data=None):
        """Hold some data for moderator approval.

        :param request_type: A `RequestType` enum value.
        :param key: The key piece of request data being held.
        :param data: Additional optional data in the form of a dictionary that
            is associated with the held request.
        :return: A unique id for this held request.
        """

    held_requests = Attribute(
        """An iterator over the held requests.

        Returned items have two attributes:
         * `id` is the held request's unique id;
         * `type` is a `RequestType` enum value.
        """)

    def of_type(request_type):
        """An iterator over the held requests of the given type.

        Returned items have two  attributes:
         * `id` is the held request's unique id;
         * `type` is a `RequestType` enum value.

         Only items with a matching `type' are returned.
         """

    def get_request(request_id, request_type):
        """Get the data associated with the request id, or None.

        :param request_id: The unique id for the request.
        :type request_id: int
        :param request_type: Optional request type that the requested id must
            match, otherwise no match is returned.
        :type request_type: `RequestType`
        :return: A 2-tuple of the key and data originally held, or None if the
            `request_id` is not in the database.
        """

    def delete_request(request_id):
        """Delete the request associated with the id.

        :param request_id: The unique id for the request.
        :raises KeyError: If `request_id` is not in the database.
        """



# Copyright (C) 2008-2012 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""The Mail-Archive.com archiver."""

from __future__ import absolute_import, unicode_literals

__metaclass__ = type
__all__ = [
    'MailArchive',
    ]


from urllib import quote
from urlparse import urljoin
from zope.interface import implements

from mailman.config import config
from mailman.interfaces.archiver import IArchiver



class MailArchive:
    """Public archiver at the Mail-Archive.com.

    Messages get archived at http://go.mail-archive.com.
    """

    implements(IArchiver)

    name = 'mail-archive'

    @staticmethod
    def list_url(mlist):
        """See `IArchiver`."""
        if mlist.archive_private:
            return None
        return urljoin(config.archiver.mail_archive.base_url,
                       quote(mlist.posting_address))

    @staticmethod
    def permalink(mlist, msg):
        """See `IArchiver`."""
        if mlist.archive_private:
            return None
        # It is the LMTP server's responsibility to ensure that the message
        # has a X-Message-ID-Hash header.  If it doesn't then there's no
        # permalink.
        message_id_hash = msg.get('x-message-id-hash')
        if message_id_hash is None:
            return None
        return urljoin(config.archiver.mail_archive.base_url, message_id_hash)

    @staticmethod
    def archive_message(mlist, msg):
        """See `IArchiver`."""
        if not mlist.archive_private:
            config.switchboards['out'].enqueue(
                msg,
                listname=mlist.fqdn_listname,
                recipients=[config.archiver.mail_archive.recipient])



# Copyright (C) 2001-2012 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""Clean up an .mbox archive file."""

import re
import sys
import mailbox
import optparse

from mailman.core.i18n import _
from mailman.version import MAILMAN_VERSION


cre = re.compile(mailbox.UnixMailbox._fromlinepattern)
# From RFC 2822, a header field name must contain only characters from 33-126
# inclusive, excluding colon.  I.e. from oct 41 to oct 176 less oct 072.  Must
# use re.match() so that it's anchored at the beginning of the line.
fre = re.compile(r'[\041-\071\073-\176]+')



def parseargs():
    parser = optparse.OptionParser(version=MAILMAN_VERSION,
                                   usage=_("""\
%prog [options] < inputfile > outputfile

The archiver looks for Unix-From lines separating messages in an mbox archive
file.  For compatibility, it specifically looks for lines that start with
'From ' -- i.e. the letters capital-F, lowercase-r, o, m, space, ignoring
everything else on the line.

Normally, any lines that start 'From ' in the body of a message should be
escaped such that a > character is actually the first on a line.  It is
possible though that body lines are not actually escaped.  This script
attempts to fix these by doing a stricter test of the Unix-From lines.  Any
lines that start From ' but do not pass this stricter test are escaped with a
'>' character."""))
    parser.add_option('-q', '--quiet',
                      default=False, action='store_true', help=_("""\
Don't print changed line information to standard error."""))
    parser.add_option('-s', '--status',
                      default=-1, type='int', help=_("""\
Print a '#' character for every n lines processed.  With a number less than or
equal to zero, suppress the '#' characters."""))
    parser.add_option('-n', '--dry-run',
                      default=False, action='store_true', help=_("""\
Don't actually output anything."""))
    opts, args = parser.parser_args()
    if args:
        parser.print_error(_('Unexpected arguments'))
    return parser, opts, args



def escape_line(line, lineno, quiet, output):
    if output:
        sys.stdout.write('>' + line)
    if not quiet:
        print >> sys.stderr, _('Unix-From line changed: $lineno')
        print >> sys.stderr, line[:-1]



def main():
    parser, opts, args = parseargs()

    lineno = 0
    statuscnt = 0
    messages = 0
    prevline = None
    while True:
        lineno += 1
        line = sys.stdin.readline()
        if not line:
            break
        if line.startswith('From '):
            if cre.match(line):
                # This is a real Unix-From line.  But it could be a message
                # /about/ Unix-From lines, so as a second order test, make
                # sure there's at least one RFC 2822 header following
                nextline = sys.stdin.readline()
                lineno += 1
                if not nextline:
                    # It was the last line of the mbox, so it couldn't have
                    # been a Unix-From
                    escape_line(line, lineno, quiet, output)
                    break
                fieldname = nextline.split(':', 1)
                if len(fieldname) < 2 or not fre.match(nextline):
                    # The following line was not a header, so this wasn't a
                    # valid Unix-From
                    escape_line(line, lineno, quiet, output)
                    if output:
                        sys.stdout.write(nextline)
                else:
                    # It's a valid Unix-From line
                    messages += 1
                    if output:
                        # Before we spit out the From_ line, make sure the
                        # previous line was blank.
                        if prevline is not None and prevline != '\n':
                            sys.stdout.write('\n')
                        sys.stdout.write(line)
                        sys.stdout.write(nextline)
            else:
                # This is a bogus Unix-From line
                escape_line(line, lineno, quiet, output)
        elif output:
            # Any old line
            sys.stdout.write(line)
        if status > 0 and (lineno % status) == 0:
            sys.stderr.write('#')
            statuscnt += 1
            if statuscnt > 50:
                print >> sys.stderr
                statuscnt = 0
        prevline = line
    print >> sys.stderr, _('%(messages)d messages found')



# Copyright (C) 2009-2012 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""Package and module utilities."""

from __future__ import absolute_import, unicode_literals

__metaclass__ = type
__all__ = [
    'call_name',
    'find_name',
    ]


import sys



def find_name(dotted_name):
    """Import and return the named object in package space.

    :param dotted_name: The dotted module path name to the object.
    :type dotted_name: string
    :return: The object.
    :rtype: object
    """
    package_path, dot, object_name = dotted_name.rpartition('.')
    __import__(package_path)
    return getattr(sys.modules[package_path], object_name)


def call_name(dotted_name, *args, **kws):
    """Imports and calls the named object in package space.

    :param dotted_name: The dotted module path name to the object.
    :type dotted_name: string
    :param args: The positional arguments.
    :type args: tuple
    :param kws: The keyword arguments.
    :type kws: dict
    :return: The object.
    :rtype: object
    """
    named_callable = find_name(dotted_name)
    return named_callable(*args, **kws)



# Copyright (C) 2010-2012 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""Module stuff."""

from __future__ import absolute_import, unicode_literals

__metaclass__ = type
__all__ = [
    'Status',
    ]


import socket

from zope.interface import implements

from mailman.bin.master import WatcherState, master_state
from mailman.core.i18n import _
from mailman.interfaces.command import ICLISubCommand



class Status:
    """Status of the Mailman system."""

    implements(ICLISubCommand)

    name = 'status'

    def add(self, parser, command_parser):
        """See `ICLISubCommand`."""
        pass

    def process(self, args):
        """See `ICLISubCommand`."""
        status, lock = master_state()
        if status is WatcherState.none:
            message = _('GNU Mailman is not running')
        elif status is WatcherState.conflict:
            hostname, pid, tempfile = lock.details
            message = _('GNU Mailman is running (master pid: $pid)')
        elif status is WatcherState.stale_lock:
            hostname, pid, tempfile = lock.details
            message =_('GNU Mailman is stopped (stale pid: $pid)')
        else:
            hostname, pid, tempfile = lock.details
            fqdn_name = socket.getfqdn()
            assert status is WatcherState.host_mismatch, (
                'Invalid enum value: %s' % status)
            message = _('GNU Mailman is in an unexpected state '
                        '($hostname != $fqdn_name)')
        print message
        return int(status)



# Copyright (C) 2007-2012 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""Application support for moderators."""

from __future__ import absolute_import, unicode_literals

__metaclass__ = type
__all__ = [
    'handle_ListDeletingEvent',
    'handle_message',
    'handle_subscription',
    'handle_unsubscription',
    'hold_message',
    'hold_subscription',
    'hold_unsubscription',
    ]


import time
import logging

from email.utils import formataddr, formatdate, getaddresses, make_msgid
from zope.component import getUtility

from mailman.app.membership import add_member, delete_member
from mailman.app.notifications import (
    send_admin_subscription_notice, send_welcome_message)
from mailman.config import config
from mailman.core.i18n import _
from mailman.email.message import UserNotification
from mailman.interfaces.action import Action
from mailman.interfaces.languages import ILanguageManager
from mailman.interfaces.listmanager import ListDeletingEvent
from mailman.interfaces.member import (
    AlreadySubscribedError, DeliveryMode, NotAMemberError)
from mailman.interfaces.messages import IMessageStore
from mailman.interfaces.requests import IListRequests, RequestType
from mailman.utilities.datetime import now
from mailman.utilities.i18n import make


NL = '\n'

vlog = logging.getLogger('mailman.vette')
slog = logging.getLogger('mailman.subscribe')



def hold_message(mlist, msg, msgdata=None, reason=None):
    """Hold a message for moderator approval.

    The message is added to the mailing list's request database.

    :param mlist: The mailing list to hold the message on.
    :param msg: The message to hold.
    :param msgdata: Optional message metadata to hold.  If not given, a new
        metadata dictionary is created and held with the message.
    :param reason: Optional string reason why the message is being held.  If
        not given, the empty string is used.
    :return: An id used to handle the held message later.
    """
    if msgdata is None:
        msgdata = {}
    else:
        # Make a copy of msgdata so that subsequent changes won't corrupt the
        # request database.  TBD: remove the `filebase' key since this will
        # not be relevant when the message is resurrected.
        msgdata = msgdata.copy()
    if reason is None:
        reason = ''
    # Add the message to the message store.  It is required to have a
    # Message-ID header.
    message_id = msg.get('message-id')
    if message_id is None:
        msg['Message-ID'] = message_id = unicode(make_msgid())
    assert isinstance(message_id, unicode), (
        'Message-ID is not a unicode: %s' % message_id)
    getUtility(IMessageStore).add(msg)
    # Prepare the message metadata with some extra information needed only by
    # the moderation interface.
    msgdata['_mod_message_id'] = message_id
    msgdata['_mod_fqdn_listname'] = mlist.fqdn_listname
    msgdata['_mod_sender'] = msg.sender
    msgdata['_mod_subject'] = msg.get('subject', _('(no subject)'))
    msgdata['_mod_reason'] = reason
    msgdata['_mod_hold_date'] = now().isoformat()
    # Now hold this request.  We'll use the message_id as the key.
    requestsdb = IListRequests(mlist)
    request_id = requestsdb.hold_request(
        RequestType.held_message, message_id, msgdata)
    return request_id



def handle_message(mlist, id, action,
                   comment=None, preserve=False, forward=None):
    message_store = getUtility(IMessageStore)
    requestdb = IListRequests(mlist)
    key, msgdata = requestdb.get_request(id)
    # Handle the action.
    rejection = None
    message_id = msgdata['_mod_message_id']
    sender = msgdata['_mod_sender']
    subject = msgdata['_mod_subject']
    if action in (Action.defer, Action.hold):
        # Nothing to do, but preserve the message for later.
        preserve = True
    elif action is Action.discard:
        rejection = 'Discarded'
    elif action is Action.reject:
        rejection = 'Refused'
        member = mlist.members.get_member(sender)
        if member:
            language = member.preferred_language
        else:
            language = None
        _refuse(mlist, _('Posting of your message titled "$subject"'),
                sender, comment or _('[No reason given]'), language)
    elif action is Action.accept:
        # Start by getting the message from the message store.
        msg = message_store.get_message_by_id(message_id)
        # Delete moderation-specific entries from the message metadata.
        for key in msgdata.keys():
            if key.startswith('_mod_'):
                del msgdata[key]
        # Add some metadata to indicate this message has now been approved.
        msgdata['approved'] = True
        msgdata['moderator_approved'] = True
        # Calculate a new filebase for the approved message, otherwise
        # delivery errors will cause duplicates.
        if 'filebase' in msgdata:
            del msgdata['filebase']
        # Queue the file for delivery.  Trying to deliver the message directly
        # here can lead to a huge delay in web turnaround.  Log the moderation
        # and add a header.
        msg['X-Mailman-Approved-At'] = formatdate(
            time.mktime(now().timetuple()), localtime=True)
        vlog.info('held message approved, message-id: %s',
                  msg.get('message-id', 'n/a'))
        # Stick the message back in the incoming queue for further
        # processing.
        config.switchboards['pipeline'].enqueue(msg, _metadata=msgdata)
    else:
        raise AssertionError('Unexpected action: {0}'.format(action))
    # Forward the message.
    if forward:
        # Get a copy of the original message from the message store.
        msg = message_store.get_message_by_id(message_id)
        # It's possible the forwarding address list is a comma separated list
        # of display_name/address pairs.
        addresses = [addr[1] for addr in getaddresses(forward)]
        language = mlist.preferred_language
        if len(addresses) == 1:
            # If the address getting the forwarded message is a member of
            # the list, we want the headers of the outer message to be
            # encoded in their language.  Otherwise it'll be the preferred
            # language of the mailing list.  This is better than sending a
            # separate message per recipient.
            member = mlist.members.get_member(addresses[0])
            if member:
                language = member.preferred_language
        with _.using(language.code):
            fmsg = UserNotification(
                addresses, mlist.bounces_address,
                _('Forward of moderated message'),
                lang=language)
        fmsg.set_type('message/rfc822')
        fmsg.attach(msg)
        fmsg.send(mlist)
    # Delete the message from the message store if it is not being preserved.
    if not preserve:
        message_store.delete_message(message_id)
        requestdb.delete_request(id)
    # Log the rejection
    if rejection:
        note = """%s: %s posting:
\tFrom: %s
\tSubject: %s"""
        if comment:
            note += '\n\tReason: ' + comment
        vlog.info(note, mlist.fqdn_listname, rejection, sender, subject)



def hold_subscription(mlist, address, display_name, password, mode, language):
    data = dict(when=now().isoformat(),
                address=address,
                display_name=display_name,
                password=password,
                delivery_mode=str(mode),
                language=language)
    # Now hold this request.  We'll use the address as the key.
    requestsdb = IListRequests(mlist)
    request_id = requestsdb.hold_request(
        RequestType.subscription, address, data)
    vlog.info('%s: held subscription request from %s',
              mlist.fqdn_listname, address)
    # Possibly notify the administrator in default list language
    if mlist.admin_immed_notify:
        subject = _(
            'New subscription request to $mlist.display_name from $address')
        text = make('subauth.txt',
                    mailing_list=mlist,
                    username=address,
                    listname=mlist.fqdn_listname,
                    admindb_url=mlist.script_url('admindb'),
                    )
        # This message should appear to come from the <list>-owner so as
        # to avoid any useless bounce processing.
        msg = UserNotification(
            mlist.owner_address, mlist.owner_address,
            subject, text, mlist.preferred_language)
        msg.send(mlist, tomoderators=True)
    return request_id



def handle_subscription(mlist, id, action, comment=None):
    requestdb = IListRequests(mlist)
    if action is Action.defer:
        # Nothing to do.
        return
    elif action is Action.discard:
        # Nothing to do except delete the request from the database.
        pass
    elif action is Action.reject:
        key, data = requestdb.get_request(id)
        _refuse(mlist, _('Subscription request'),
                data['address'],
                comment or _('[No reason given]'),
                lang=getUtility(ILanguageManager)[data['language']])
    elif action is Action.accept:
        key, data = requestdb.get_request(id)
        enum_value = data['delivery_mode'].split('.')[-1]
        delivery_mode = DeliveryMode(enum_value)
        address = data['address']
        display_name = data['display_name']
        language = getUtility(ILanguageManager)[data['language']]
        password = data['password']
        try:
            add_member(mlist, address, display_name, password,
                       delivery_mode, language)
        except AlreadySubscribedError:
            # The address got subscribed in some other way after the original
            # request was made and accepted.
            pass
        else:
            if mlist.send_welcome_message:
                send_welcome_message(mlist, address, language, delivery_mode)
            if mlist.admin_notify_mchanges:
                send_admin_subscription_notice(
                    mlist, address, display_name, language)
        slog.info('%s: new %s, %s %s', mlist.fqdn_listname,
                  delivery_mode, formataddr((display_name, address)),
                  'via admin approval')
    else:
        raise AssertionError('Unexpected action: {0}'.format(action))
    # Delete the request from the database.
    requestdb.delete_request(id)



def hold_unsubscription(mlist, address):
    data = dict(address=address)
    requestsdb = IListRequests(mlist)
    request_id = requestsdb.hold_request(
        RequestType.unsubscription, address, data)
    vlog.info('%s: held unsubscription request from %s',
              mlist.fqdn_listname, address)
    # Possibly notify the administrator of the hold
    if mlist.admin_immed_notify:
        subject = _(
            'New unsubscription request from $mlist.display_name by $address')
        text = make('unsubauth.txt',
                    mailing_list=mlist,
                    address=address,
                    listname=mlist.fqdn_listname,
                    admindb_url=mlist.script_url('admindb'),
                    )
        # This message should appear to come from the <list>-owner so as
        # to avoid any useless bounce processing.
        msg = UserNotification(
            mlist.owner_address, mlist.owner_address,
            subject, text, mlist.preferred_language)
        msg.send(mlist, tomoderators=True)
    return request_id



def handle_unsubscription(mlist, id, action, comment=None):
    requestdb = IListRequests(mlist)
    key, data = requestdb.get_request(id)
    address = data['address']
    if action is Action.defer:
        # Nothing to do.
        return
    elif action is Action.discard:
        # Nothing to do except delete the request from the database.
        pass
    elif action is Action.reject:
        key, data = requestdb.get_request(id)
        _refuse(mlist, _('Unsubscription request'), address,
                comment or _('[No reason given]'))
    elif action is Action.accept:
        key, data = requestdb.get_request(id)
        try:
            delete_member(mlist, address)
        except NotAMemberError:
            # User has already been unsubscribed.
            pass
        slog.info('%s: deleted %s', mlist.fqdn_listname, address)
    else:
        raise AssertionError('Unexpected action: {0}'.format(action))
    # Delete the request from the database.
    requestdb.delete_request(id)



def _refuse(mlist, request, recip, comment, origmsg=None, lang=None):
    # As this message is going to the requester, try to set the language to
    # his/her language choice, if they are a member.  Otherwise use the list's
    # preferred language.
    display_name = mlist.display_name
    if lang is None:
        member = mlist.members.get_member(recip)
        lang = (mlist.preferred_language
                if member is None
                else member.preferred_language)
    text = make('refuse.txt',
                mailing_list=mlist,
                language=lang.code,
                listname=mlist.fqdn_listname,
                request=request,
                reason=comment,
                adminaddr=mlist.owner_address,
                )
    with _.using(lang.code):
        # add in original message, but not wrap/filled
        if origmsg:
            text = NL.join(
                [text,
                 '---------- ' + _('Original Message') + ' ----------',
                 str(origmsg)
                 ])
        subject = _('Request to mailing list "$display_name" rejected')
    msg = UserNotification(recip, mlist.bounces_address, subject, text, lang)
    msg.send(mlist)



def handle_ListDeletingEvent(event):
    if not isinstance(event, ListDeletingEvent):
        return
    # Get the held requests database for the mailing list.  Since the mailing
    # list is about to get deleted, we can delete all associated requests.
    requestsdb = IListRequests(event.mailing_list)
    for request in requestsdb.held_requests:
        requestsdb.delete_request(request.id)



# Copyright (C) 2010-2012 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""Test master watcher utilities."""

from __future__ import absolute_import, unicode_literals

__metaclass__ = type
__all__ = [
    ]


import os
import errno
import tempfile
import unittest

from flufl.lock import Lock

from mailman.bin import master



class TestMasterLock(unittest.TestCase):
    def setUp(self):
        fd, self.lock_file = tempfile.mkstemp()
        os.close(fd)
        # The lock file should not exist before we try to acquire it.
        os.remove(self.lock_file)

    def tearDown(self):
        # Unlocking removes the lock file, but just to be safe (i.e. in case
        # of errors).
        try:
            os.remove(self.lock_file)
        except OSError as error:
            if error.errno != errno.ENOENT:
                raise

    def test_acquire_lock_1(self):
        lock = master.acquire_lock_1(False, self.lock_file)
        is_locked = lock.is_locked
        lock.unlock()
        self.failUnless(is_locked)

    def test_master_state(self):
        my_lock = Lock(self.lock_file)
        # Mailman is not running.
        state, lock = master.master_state(self.lock_file)
        self.assertEqual(state, master.WatcherState.none)
        # Acquire the lock as if another process had already started the
        # master.
        my_lock.lock()
        try:
            state, lock = master.master_state(self.lock_file)
        finally:
            my_lock.unlock()
        self.assertEqual(state, master.WatcherState.conflict)
        # XXX test stale_lock and host_mismatch states.



# Copyright (C) 2012 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""Template loader."""

from __future__ import absolute_import, print_function, unicode_literals

__metaclass__ = type
__all__ = [
    'TemplateLoader',
    ]


import urllib2

from contextlib import closing
from urllib import addinfourl
from urlparse import urlparse
from zope.component import getUtility
from zope.interface import implements

from mailman.utilities.i18n import TemplateNotFoundError, find
from mailman.interfaces.languages import ILanguageManager
from mailman.interfaces.listmanager import IListManager
from mailman.interfaces.templates import ITemplateLoader



class MailmanHandler(urllib2.BaseHandler):
    # Handle internal mailman: URLs.
    def mailman_open(self, req):
        # Parse urls of the form:
        #
        # mailman:///<fqdn_listname>/<language>/<template_name>
        #
        # where only the template name is required.
        mlist = code = template = None
        # Parse the full requested URL and be sure it's something we handle.
        original_url = req.get_full_url()
        parsed = urlparse(original_url)
        assert parsed.scheme == 'mailman'
        # The path can contain one, two, or three components.  Since no empty
        # path components are legal, filter them out.
        parts = filter(None, parsed.path.split('/'))
        if len(parts) == 0:
            raise urllib2.URLError('No template specified')
        elif len(parts) == 1:
            template = parts[0]
        elif len(parts) == 2:
            part0, template = parts
            # Is part0 a language code or a mailing list?  It better be one or
            # the other, and there's no possibility of namespace collisions
            # because language codes don't contain @ and mailing list names
            # MUST contain @.
            language = getUtility(ILanguageManager).get(part0)
            mlist = getUtility(IListManager).get(part0)
            if language is None and mlist is None:
                raise urllib2.URLError('Bad language or list name')
            elif mlist is None:
                code = language.code
        elif len(parts) == 3:
            fqdn_listname, code, template = parts
            mlist = getUtility(IListManager).get(fqdn_listname)
            if mlist is None:
                raise urllib2.URLError('Missing list')
            language = getUtility(ILanguageManager).get(code)
            if language is None:
                raise urllib2.URLError('No such language')
            code = language.code
        else:
            raise urllib2.URLError('No such file')
        # Find the template, mutating any missing template exception.
        try:
            path, fp = find(template, mlist, code)
        except TemplateNotFoundError:
            raise urllib2.URLError('No such file')
        return addinfourl(fp, {}, original_url)



class TemplateLoader:
    """Loader of templates, with caching and support for mailman:// URIs."""

    implements(ITemplateLoader)

    def __init__(self):
        opener = urllib2.build_opener(MailmanHandler())
        urllib2.install_opener(opener)

    def get(self, uri):
        """See `ITemplateLoader`."""
        with closing(urllib2.urlopen(uri)) as fp:
            return fp.read()



# Copyright (C) 2009-2012 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""One last digest."""

from __future__ import absolute_import, unicode_literals

__metaclass__ = type
__all__ = [
    'OneLastDigest',
    ]


from storm.locals import Int, Reference
from zope.interface import implements

from mailman.database.model import Model
from mailman.database.types import Enum
from mailman.interfaces.digests import IOneLastDigest
from mailman.interfaces.member import DeliveryMode



class OneLastDigest(Model):
    implements(IOneLastDigest)

    id = Int(primary=True)

    mailing_list_id = Int()
    mailing_list = Reference(mailing_list_id, 'MailingList.id')

    address_id = Int()
    address = Reference(address_id, 'Address.id')

    delivery_mode = Enum(DeliveryMode)

    def __init__(self, mailing_list, address, delivery_mode):
        self.mailing_list = mailing_list
        self.address = address
        self.delivery_mode = delivery_mode



# Copyright (C) 2007-2012 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""Various rule helpers"""

from __future__ import absolute_import, unicode_literals

__metaclass__ = type
__all__ = [
    'initialize',
    ]


from zope.interface.verify import verifyObject

from mailman.app.finder import find_components
from mailman.config import config
from mailman.interfaces.rules import IRule



def initialize():
    """Find and register all rules in all plugins."""
    # Find rules in plugins.
    for rule_class in find_components('mailman.rules', IRule):
        rule = rule_class()
        verifyObject(IRule, rule)
        assert rule.name not in config.rules, (
            'Duplicate rule "{0}" found in {1}'.format(
                rule.name, rule_class))
        config.rules[rule.name] = rule



# Copyright (C) 2008-2012 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""MHonArc archiver."""

from __future__ import absolute_import, unicode_literals

__metaclass__ = type
__all__ = [
    'MHonArc',
    ]


import logging
import subprocess

from urlparse import urljoin
from zope.interface import implements

from mailman.config import config
from mailman.interfaces.archiver import IArchiver
from mailman.utilities.string import expand


log = logging.getLogger('mailman.archiver')



class MHonArc:
    """Local MHonArc archiver."""

    implements(IArchiver)

    name = 'mhonarc'

    @staticmethod
    def list_url(mlist):
        """See `IArchiver`."""
        # XXX What about private MHonArc archives?
        return expand(config.archiver.mhonarc.base_url,
                      dict(listname=mlist.fqdn_listname,
                           hostname=mlist.domain.url_host,
                           fqdn_listname=mlist.fqdn_listname,
                           ))

    @staticmethod
    def permalink(mlist, msg):
        """See `IArchiver`."""
        # XXX What about private MHonArc archives?
        # It is the LMTP server's responsibility to ensure that the message
        # has a X-Message-ID-Hash header.  If it doesn't then there's no
        # permalink.
        message_id_hash = msg.get('x-message-id-hash')
        if message_id_hash is None:
            return None
        return urljoin(MHonArc.list_url(mlist), message_id_hash)

    @staticmethod
    def archive_message(mlist, msg):
        """See `IArchiver`."""
        substitutions = config.__dict__.copy()
        substitutions['listname'] = mlist.fqdn_listname
        command = expand(config.archiver.mhonarc.command, substitutions)
        proc = subprocess.Popen(
            command, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
            shell=True)
        stdout, stderr = proc.communicate(msg.as_string())
        if proc.returncode != 0:
            log.error('%s: mhonarc subprocess had non-zero exit code: %s' %
                      (msg['message-id'], proc.returncode))
        log.info(stdout)
        log.error(stderr)



# Copyright (C) 2011-2012 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""i18n template search and interpolation."""

from __future__ import absolute_import, print_function, unicode_literals

__metaclass__ = type
__all__ = [
    'TemplateNotFoundError',
    'find',
    'make',
    'search',
    ]


import os
import sys
import errno

from itertools import product
from pkg_resources import resource_filename

from mailman.config import config
from mailman.core.constants import system_preferences
from mailman.core.errors import MailmanException
from mailman.core.i18n import _
from mailman.utilities.string import expand, wrap as wrap_text



class TemplateNotFoundError(MailmanException):
    """The named template was not found."""

    def __init__(self, template_file):
        self.template_file = template_file

    def __str__(self):
        return self.template_file



def search(template_file, mlist=None, language=None):
    """Generator that provides file system search order.

    This is Mailman's internal template search algorithm.  The first locations
    searched are within the $template_dir directory, allowing a site to
    override a template for a specific mailing list, all the mailing lists in
    a domain, or site-wide.

    The <language> path component is variable, and described below.

    * The list-specific language directory
      $template_dir/lists/<mlist.fqdn_listname>/<language>

    * The domain-specific language directory
      $template_dir/domains/<mlist.mail_host>/<language>

    * The site-wide language directory
      $template_dir/site/<language>

    The <language> path component is calculated as follows, in this order:

    * The `language` parameter if given
    * `mlist.preferred_language` if given
    * The server's default language
    * English ('en')

    Languages are iterated after each of the four locations are searched.  So
    for example, when searching for the 'foo.txt' template, where the server's
    default language is 'fr', the mailing list's (test@example.com) language
    is 'de' and the `language` parameter is 'it', these locations are searched
    in order:

    * $template_dir/lists/test@example.com/it/foo.txt
    * $template_dir/domains/example.com/it/foo.txt
    * $template_dir/site/it/foo.txt

    * $template_dir/lists/test@example.com/de/foo.txt
    * $template_dir/domains/example.com/de/foo.txt
    * $template_dir/site/de/foo.txt

    * $template_dir/lists/test@example.com/fr/foo.txt
    * $template_dir/domains/example.com/fr/foo.txt
    * $template_dir/site/fr/foo.txt

    * $template_dir/lists/test@example.com/en/foo.txt
    * $template_dir/domains/example.com/en/foo.txt
    * $template_dir/site/en/foo.txt

    After all those paths are searched, the final fallback is the English
    template within the Mailman source tree.

    * <source_dir>/templates/en/foo.txt
    """
    # The languages in search order.
    languages = ['en', system_preferences.preferred_language.code]
    if mlist is not None:
        languages.append(mlist.preferred_language.code)
    if language is not None:
        languages.append(language)
    languages.reverse()
    # The non-language qualified $template_dir paths in search order.
    paths = [os.path.join(config.TEMPLATE_DIR, 'site')]
    if mlist is not None:
        paths.append(os.path.join(
            config.TEMPLATE_DIR, 'domains', mlist.mail_host))
        paths.append(os.path.join(
            config.TEMPLATE_DIR, 'lists', mlist.fqdn_listname))
    paths.reverse()
    for language, path in product(languages, paths):
        yield os.path.join(path, language, template_file)
    # Finally, fallback to the in-tree English template.
    templates_dir = resource_filename('mailman', 'templates')
    yield os.path.join(templates_dir, 'en', template_file)



def find(template_file, mlist=None, language=None, _trace=False):
    """Use Mailman's internal template search order to find a template.

    :param template_file: The name of the template file to search for.
    :type template_file: string
    :param mlist: Optional mailing list used as the context for
        searching for the template file.  The list's preferred language will
        influence the search, as will the list's data directory.
    :type mlist: `IMailingList`
    :param language: Optional language code, which influences the search.
    :type language: string
    :param _trace: Enable printing of debugging information during
        template search.
    :type _trace: bool
    :return: A tuple of the file system path to the first matching template,
        and an open file object allowing reading of the file.
    :rtype: (string, file)
    :raises TemplateNotFoundError: when the template could not be found.
    """
    raw_search_order = search(template_file, mlist, language)
    for path in raw_search_order:
        try:
            if _trace:
                print('@@@', path, end='', file=sys.stderr)
            fp = open(path)
        except IOError as error:
            if error.errno == errno.ENOENT:
                if _trace:
                    print('MISSING', file=sys.stderr)
            else:
                raise
        else:
            if _trace:
                print('FOUND:', path, file=sys.stderr)
            return path, fp
    raise TemplateNotFoundError(template_file)


def make(template_file, mlist=None, language=None, wrap=True,
         _trace=False, **kw):
    """Locate and 'make' a template file.

    The template file is located as with `find()`, and the resulting text is
    optionally wrapped and interpolated with the keyword argument dictionary.

    :param template_file: The name of the template file to search for.
    :type template_file: string
    :param mlist: Optional mailing list used as the context for
        searching for the template file.  The list's preferred language will
        influence the search, as will the list's data directory.
    :type mlist: `IMailingList`
    :param language: Optional language code, which influences the search.
    :type language: string
    :param wrap: When True, wrap the text.
    :type wrap: bool
    :param _trace: Passed through to ``find()``, this enables printing of
        debugging information during template search.
    :type _trace: bool
    :param **kw: Keyword arguments for template interpolation.
    :return: The interpolated text.
    :rtype: string
    :raises TemplateNotFoundError: when the template could not be found.
    """
    path, fp = find(template_file, mlist, language, _trace)
    try:
        # XXX Removing the trailing newline is a hack carried over from
        # Mailman 2.  The (stripped) template text is then passed through the
        # translation catalog.  This ensures that the translated text is
        # unicode, and also allows for volunteers to translate the templates
        # into the language catalogs.
        template = _(fp.read()[:-1])
    finally:
        fp.close()
    assert isinstance(template, unicode), 'Translated template is not unicode'
    text = expand(template, kw)
    if wrap:
        return wrap_text(text)
    return text



# Copyright (C) 2002-2012 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""Test bounce detection on message files."""


from __future__ import absolute_import, unicode_literals

__metaclass__ = type
__all__ = [
    'main',
    ]


import sys

from mailman.core.i18n import _
from mailman.options import Options



class ScriptOptions(Options):
    """Options for onebounce."""

    usage = _("""\
%prog [options]

Test the bounce detection for message files.""")

    def add_options(self):
        """See `Options`."""
        self.parser.add_option(
            '-a', '--all',
            default=False, action='store_true',
            help=_("""\
Run the message through all the registered bounce modules.  Normally this
script stops at the first match."""))
        self.parser.add_option(
            '-m', '--module',
            type='string', help=_("""
Run the message through just the named bounce module."""))
        self.parser.add_option(
            '-l', '--list',
            default=False, action='store_true',
            help=_('List all available bounce modules and exit.'))
        self.parser.add_option(
            '-v', '--verbose',
            default=False, action='store_true',
            help=_('Increase verbosity.'))



def main():
    """bin/onebounce"""
    options = ScriptOptions()
    options.initialize()

    if options.options.list:
        print 'list of available bounce modules.'
        sys.exit(0)



# Copyright (C) 2008-2012 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""Initialize the email commands."""

from __future__ import unicode_literals

__metaclass__ = type
__all__ = [
    'initialize',
    ]


from zope.interface.verify import verifyObject

from mailman.app.finder import find_components
from mailman.config import config
from mailman.interfaces.command import IEmailCommand



def initialize():
    """Initialize the email commands."""
    for command_class in find_components('mailman.commands', IEmailCommand):
        command = command_class()
        verifyObject(IEmailCommand, command)
        assert command_class.name not in config.commands, (
            'Duplicate email command "{0}" found in {1}'.format(
                command_class.name, command_class.__module__))
        config.commands[command_class.name] = command_class()



# Copyright (C) 2007-2012 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

import csv
import optparse

from zope.component import getUtility

from mailman import Utils
from mailman import passwords
from mailman.app.membership import add_member
from mailman.app.notifications import (
    send_admin_subscription_notice, send_welcome_message)
from mailman.configuration import config
from mailman.core.i18n import _
from mailman.initialize import initialize
from mailman.interfaces.members import DeliveryMode
from mailman.interfaces.usermanager import IUserManager
from mailman.version import MAILMAN_VERSION


DELIVERY_MODES = {
    'regular':  DeliveryMode.regular,
    'plain':    DeliveryMode.plaintext_digests,
    'mime':     DeliveryMode.mime_digests,
    }



def parseargs():
    parser = optparse.OptionParser(version=MAILMAN_VERSION,
                                   usage=_("""\
%prog [options] csv-file

Set the membership of a mailing list to that described in a CSV file.  Each
row of the CSV file has the following format.  Only the address column is
required.

    - email address
    - full name (default: the empty string)
    - delivery mode (default: regular delivery) [1]

[1] The delivery mode is a case insensitive string of the following values:

    regular     - regular, i.e. immediate delivery
    mime        - MIME digest delivery
    plain       - plain text (RFC 1153) digest delivery

Any address not included in the CSV file is removed from the list membership.
"""))
    parser.add_option('-l', '--listname',
                      type='string', help=_("""\
Mailng list to set the membership for."""))
    parser.add_option('-w', '--welcome-msg',
                      type='string', metavar='<y|n>', help=_("""\
Set whether or not to send the list members a welcome message, overriding
whatever the list's 'send_welcome_msg' setting is."""))
    parser.add_option('-a', '--admin-notify',
                      type='string', metavar='<y|n>', help=_("""\
Set whether or not to send the list administrators a notification on the
success/failure of these subscriptions, overriding whatever the list's
'admin_notify_mchanges' setting is."""))
    parser.add_option('-v', '--verbose', action='store_true',
                      help=_('Increase verbosity'))
    parser.add_option('-C', '--config',
                      help=_('Alternative configuration file to use'))
    opts, args = parser.parse_args()
    if opts.welcome_msg is not None:
        ch = opts.welcome_msg[0].lower()
        if ch == 'y':
            opts.welcome_msg = True
        elif ch == 'n':
            opts.welcome_msg = False
        else:
            parser.error(_('Illegal value for -w: $opts.welcome_msg'))
    if opts.admin_notify is not None:
        ch = opts.admin_notify[0].lower()
        if ch == 'y':
            opts.admin_notify = True
        elif ch == 'n':
            opts.admin_notify = False
        else:
            parser.error(_('Illegal value for -a: $opts.admin_notify'))
    return parser, opts, args



def parse_file(filename):
    members = {}
    with open(filename) as fp:
        for row in csv.reader(fp):
            if len(row) == 0:
                continue
            elif len(row) == 1:
                address = row[0]
                real_name = None
                delivery_mode = DeliveryMode.regular
            elif len(row) == 2:
                address, real_name = row
                delivery_mode = DeliveryMode.regular
            else:
                # Ignore extra columns
                address, real_name = row[0:2]
                delivery_mode = DELIVERY_MODES.get(row[2].lower())
                if delivery_mode is None:
                    delivery_mode = DeliveryMode.regular
            members[address] = real_name, delivery_mode
    return members



def main():
    parser, opts, args = parseargs()
    initialize(opts.config)

    mlist = config.db.list_manager.get(opts.listname)
    if mlist is None:
        parser.error(_('No such list: $opts.listname'))

    # Set up defaults.
    if opts.welcome_msg is None:
        send_welcome_msg = mlist.send_welcome_msg
    else:
        send_welcome_msg = opts.welcome_msg
    if opts.admin_notify is None:
        admin_notify = mlist.admin_notify_mchanges
    else:
        admin_notify = opts.admin_notify

    # Parse the csv files.
    member_data = {}
    for filename in args:
        member_data.update(parse_file(filename))

    future_members = set(member_data)
    current_members = set(obj.address for obj in mlist.members.addresses)
    add_members = future_members - current_members
    delete_members = current_members - future_members
    change_members = current_members & future_members
    
    with _.using(mlist.preferred_language):
        # Start by removing all the delete members.
        for address in delete_members:
            print _('deleting address: $address')
            member = mlist.members.get_member(address)
            member.unsubscribe()
        # For all members that are in both lists, update their full name and
        # delivery mode.
        for address in change_members:
            print _('updating address: $address')
            real_name, delivery_mode = member_data[address]
            member = mlist.members.get_member(address)
            member.preferences.delivery_mode = delivery_mode
            user = getUtility(IUserManager).get_user(address)
            user.real_name = real_name
        for address in add_members:
            print _('adding address: $address')
            real_name, delivery_mode = member_data[address]
            password = passwords.make_secret(
                Utils.MakeRandomPassword(),
                passwords.lookup_scheme(config.PASSWORD_SCHEME))
            add_member(mlist, address, real_name, password, delivery_mode,
                       mlist.preferred_language, send_welcome_msg,
                       admin_notify)
            if send_welcome_msg:
                send_welcome_message(mlist, address, language, delivery_mode)
            if admin_notify:
                send_admin_subscription_notice(mlist, address, real_name)

    config.db.flush()



if __name__ == '__main__':
    main()



# Copyright (C) 2009-2012 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""The 'members' subcommand."""

from __future__ import absolute_import, print_function, unicode_literals

__metaclass__ = type
__all__ = [
    'Members',
    ]


import sys
import codecs

from email.utils import formataddr, parseaddr
from flufl.password import generate
from operator import attrgetter
from zope.component import getUtility
from zope.interface import implements

from mailman.app.membership import add_member
from mailman.config import config
from mailman.core.i18n import _
from mailman.interfaces.command import ICLISubCommand
from mailman.interfaces.listmanager import IListManager
from mailman.interfaces.member import (
    AlreadySubscribedError, DeliveryMode, DeliveryStatus)



class Members:
    """Manage list memberships.  With no arguments, list all members."""

    implements(ICLISubCommand)

    name = 'members'

    def add(self, parser, command_parser):
        """See `ICLISubCommand`."""
        self.parser = parser
        command_parser.add_argument(
            '-a', '--add',
            dest='input_filename', metavar='FILENAME',
            help=_("""\
            Add all member addresses in FILENAME.  FILENAME can be '-' to
            indicate standard input.  Blank lines and lines That start with a
            '#' are ignored.  Without this option, this command displays
            mailing list members."""))
        command_parser.add_argument(
            '-o', '--output',
            dest='output_filename', metavar='FILENAME',
            help=_("""Display output to FILENAME instead of stdout.  FILENAME
            can be '-' to indicate standard output."""))
        command_parser.add_argument(
            '-r', '--regular',
            default=None, action='store_true',
            help=_('Display only regular delivery members.'))
        command_parser.add_argument(
            '-d', '--digest',
            default=None, metavar='KIND',
            # BAW 2010-01-23 summary digests are not really supported yet.
            choices=('any', 'plaintext', 'mime'),
            help=_("""Display only digest members of KIND.  'any' means any
            digest type, 'plaintext' means only plain text (RFC 1153) type
            digests, 'mime' means MIME type digests."""))
        command_parser.add_argument(
            '-n', '--nomail',
            default=None, metavar='WHY',
            choices=('enabled', 'any', 'unknown'
                     'byadmin', 'byuser', 'bybounces'),
            help=_("""Display only members with a given delivery
            status. 'enabled' means all members whose delivery is enabled,
            'any' means members whose delivery is disabled for any reason,
            'byuser' means that the member disabled their own delivery,
            'bybounces' means that delivery was disabled by the automated
            bounce processor, 'byadmin' means delivery was disabled by the
            list administrator or moderator, and 'unknown' means that delivery
            was disabled for unknown (legacy) reasons."""))
        # Required positional argument.
        command_parser.add_argument(
            'listname', metavar='LISTNAME', nargs=1,
            help=_("""\
            The 'fully qualified list name', i.e. the posting address of the
            mailing list.  It must be a valid email address and the domain
            must be registered with Mailman.  List names are forced to lower
            case."""))

    def process(self, args):
        """See `ICLISubCommand`."""
        assert len(args.listname) == 1, 'Missing mailing list name'
        fqdn_listname = args.listname[0]
        mlist = getUtility(IListManager).get(fqdn_listname)
        if mlist is None:
            self.parser.error(_('No such list: $fqdn_listname'))
        if args.input_filename is None:
            self.display_members(mlist, args)
        else:
            self.add_members(mlist, args)

    def display_members(self, mlist, args):
        """Display the members of a mailing list.

        :param mlist: The mailing list to operate on.
        :type mlist: `IMailingList`
        :param args: The command line arguments.
        :type args: `argparse.Namespace`
        """
        if args.output_filename == '-' or args.output_filename is None:
            fp = sys.stdout
        else:
            fp = codecs.open(args.output_filename, 'w', 'utf-8')
        if args.digest == 'any':
            digest_types = [DeliveryMode.plaintext_digests,
                            DeliveryMode.mime_digests,
                            DeliveryMode.summary_digests]
        elif args.digest is not None:
            digest_types = [DeliveryMode(args.digest + '_digests')]
        else:
            # Don't filter on digest type.
            pass
        if args.nomail is None:
            # Don't filter on delivery status.
            pass
        elif args.nomail == 'byadmin':
            status_types = [DeliveryStatus.by_moderator]
        elif args.nomail.startswith('by'):
            status_types = [DeliveryStatus('by_' + args.nomail[2:])]
        elif args.nomail == 'enabled':
            status_types = [DeliveryStatus.enabled]
        elif args.nomail == 'unknown':
            status_types = [DeliveryStatus.unknown]
        elif args.nomail == 'any':
            status_types = [DeliveryStatus.by_user,
                            DeliveryStatus.by_bounces,
                            DeliveryStatus.by_moderator,
                            DeliveryStatus.unknown]
        else:
            raise AssertionError('Unknown delivery status: %s' % args.nomail)
        try:
            addresses = list(mlist.members.addresses)
            if len(addresses) == 0:
                print(mlist.fqdn_listname, 'has no members', file=fp)
                return
            for address in sorted(addresses, key=attrgetter('email')):
                if args.regular:
                    member = mlist.members.get_member(address.email)
                    if member.delivery_mode != DeliveryMode.regular:
                        continue
                if args.digest is not None:
                    member = mlist.members.get_member(address.email)
                    if member.delivery_mode not in digest_types:
                        continue
                if args.nomail is not None:
                    member = mlist.members.get_member(address.email)
                    if member.delivery_status not in status_types:
                        continue
                print(
                    formataddr((address.display_name, address.original_email)),
                    file=fp)
        finally:
            if fp is not sys.stdout:
                fp.close()

    def add_members(self, mlist, args):
        """Add the members in a file to a mailing list.

        :param mlist: The mailing list to operate on.
        :type mlist: `IMailingList`
        :param args: The command line arguments.
        :type args: `argparse.Namespace`
        """
        if args.input_filename == '-':
            fp = sys.stdin
        else:
            fp = codecs.open(args.input_filename, 'r', 'utf-8')
        try:
            for line in fp:
                # Ignore blank lines and lines that start with a '#'.
                if line.startswith('#') or len(line.strip()) == 0:
                    continue
                # Parse the line and ensure that the values are unicodes.
                display_name, email = parseaddr(line)
                display_name = display_name.decode(fp.encoding)
                email = email.decode(fp.encoding)
                # Give the user a default, user-friendly password.
                password = generate(int(config.passwords.password_length))
                try:
                    add_member(mlist, email, display_name, password,
                               DeliveryMode.regular,
                               mlist.preferred_language.code)
                except AlreadySubscribedError:
                    # It's okay if the address is already subscribed, just
                    # print a warning and continue.
                    print('Already subscribed (skipping):', 
                          email, display_name)
        finally:
            if fp is not sys.stdin:
                fp.close()
        config.db.commit()



# Copyright (C) 2007-2012 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""Interface describing the basics of a member."""

from __future__ import absolute_import, unicode_literals

__metaclass__ = type
__all__ = [
    'AlreadySubscribedError',
    'DeliveryMode',
    'DeliveryStatus',
    'IMember',
    'MemberRole',
    'MembershipError',
    'MembershipIsBannedError',
    'MissingPreferredAddressError',
    'NotAMemberError',
    ]


from flufl.enum import Enum
from zope.interface import Interface, Attribute

from mailman.core.errors import MailmanError



class DeliveryMode(Enum):
    # Regular (i.e. non-digest) delivery
    regular = 1
    # Plain text digest delivery
    plaintext_digests = 2
    # MIME digest delivery
    mime_digests = 3
    # Summary digests
    summary_digests = 4



class DeliveryStatus(Enum):
    # Delivery is enabled
    enabled = 1
    # Delivery was disabled by the user
    by_user = 2
    # Delivery was disabled due to bouncing addresses
    by_bounces = 3
    # Delivery was disabled by an administrator or moderator
    by_moderator = 4
    # Disabled for unknown reasons.
    unknown = 5



class MemberRole(Enum):
    member = 1
    owner = 2
    moderator = 3
    nonmember = 4



class MembershipError(MailmanError):
    """Base exception for all membership errors."""


class AlreadySubscribedError(MembershipError):
    """The member is already subscribed to the mailing list with this role."""

    def __init__(self, fqdn_listname, email, role):
        super(AlreadySubscribedError, self).__init__()
        self.fqdn_listname = fqdn_listname
        self.email = email
        self.role = role

    def __str__(self):
        return '{0} is already a {1} of mailing list {2}'.format(
            self.email, self.role, self.fqdn_listname)


class MembershipIsBannedError(MembershipError):
    """The address is not allowed to subscribe to the mailing list."""

    def __init__(self, mlist, address):
        super(MembershipIsBannedError, self).__init__()
        self._mlist = mlist
        self._address = address

    def __str__(self):
        return '{0} is not allowed to subscribe to {1.fqdn_listname}'.format(
            self._address, self._mlist)


class MissingPreferredAddressError(MembershipError):
    """A user without a preferred address attempted to subscribe."""

    def __init__(self, user):
        super(MissingPreferredAddressError, self).__init__()
        self._user = user

    def __str__(self):
        return 'User must have a preferred address: {0}'.format(self._user)


class NotAMemberError(MembershipError):
    """The address is not a member of the mailing list."""

    def __init__(self, mlist, address):
        super(NotAMemberError, self).__init__()
        self._mlist = mlist
        self._address = address

    def __str__(self):
        return '{0} is not a member of {1.fqdn_listname}'.format(
            self._address, self._mlist)



class IMember(Interface):
    """A member of a mailing list."""

    member_id = Attribute(
        """The member's unique, random identifier as a UUID.""")

    mailing_list = Attribute(
        """The mailing list subscribed to.""")

    address = Attribute(
        """The email address that's subscribed to the list.""")

    user = Attribute(
        """The user associated with this member.""")

    preferences = Attribute(
        """This member's preferences.""")

    role = Attribute(
        """The role of this membership.""")

    moderation_action = Attribute(
        """The moderation action for this member as an `Action`.""")

    def unsubscribe():
        """Unsubscribe (and delete) this member from the mailing list."""

    acknowledge_posts = Attribute(
        """Send an acknowledgment for every posting?

        Unlike going through the preferences, this attribute return the
        preference value based on the following lookup order:

        1. The member
        2. The address
        3. The user
        4. System default
        """)

    preferred_language = Attribute(
        """The preferred language for interacting with a mailing list.

        Unlike going through the preferences, this attribute return the
        preference value based on the following lookup order:

        1. The member
        2. The address
        3. The user
        4. System default
        """)

    receive_list_copy = Attribute(
        """Should an explicit recipient receive a list copy?

        Unlike going through `preferences`, this attribute returns the
        preference value based on the following lookup order:

        1. The member
        2. The address
        3. The user
        4. System default
        """)

    receive_own_postings = Attribute(
        """Should the poster get a list copy of their own messages?

        Unlike going through `preferences`, this attribute returns the
        preference value based on the following lookup order:

        1. The member
        2. The address
        3. The user
        4. System default
        """)

    delivery_mode = Attribute(
        """The preferred delivery mode.

        Unlike going through `preferences`, this attribute returns the
        preference value based on the following lookup order:

        1. The member
        2. The address
        3. The user
        4. System default
        """)

    delivery_status = Attribute(
        """The delivery status.

        Unlike going through `preferences`, this attribute returns the
        preference value based on the following lookup order:

        1. The member
        2. The address
        3. The user
        4. System default

        XXX I'm not sure this is the right place to put this.""")

    options_url = Attribute(
        """Return the url for the given member's option page.

        XXX This needs a serious re-think in the face of the unified user
        database, since a member's options aren't tied to any specific mailing
        list.  So in what part of the web-space does the user's options live?
        """)



# Copyright (C) 2011-2012 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""Test the ListManager."""

from __future__ import absolute_import, unicode_literals

__metaclass__ = type
__all__ = [
    ]


import unittest

from zope.component import getUtility

from mailman.app.lifecycle import create_list
from mailman.app.moderator import hold_message
from mailman.interfaces.listmanager import (
    IListManager, ListCreatedEvent, ListCreatingEvent, ListDeletedEvent,
    ListDeletingEvent)
from mailman.interfaces.messages import IMessageStore
from mailman.interfaces.requests import IListRequests
from mailman.interfaces.subscriptions import ISubscriptionService
from mailman.interfaces.usermanager import IUserManager
from mailman.testing.helpers import (
    event_subscribers, specialized_message_from_string)
from mailman.testing.layers import ConfigLayer



class TestListManager(unittest.TestCase):
    layer = ConfigLayer

    def setUp(self):
        self._events = []

    def _record_event(self, event):
        self._events.append(event)

    def test_create_list_event(self):
        # Test that creating a list in the list manager propagates the
        # expected events.
        with event_subscribers(self._record_event):
            mlist = getUtility(IListManager).create('test@example.com')
        self.assertEqual(len(self._events), 2)
        self.assertTrue(isinstance(self._events[0], ListCreatingEvent))
        self.assertEqual(self._events[0].fqdn_listname, 'test@example.com')
        self.assertTrue(isinstance(self._events[1], ListCreatedEvent))
        self.assertEqual(self._events[1].mailing_list, mlist)

    def test_delete_list_event(self):
        # Test that deleting a list in the list manager propagates the
        # expected event.
        mlist = create_list('another@example.com')
        with event_subscribers(self._record_event):
            getUtility(IListManager).delete(mlist)
        self.assertEqual(len(self._events), 2)
        self.assertTrue(isinstance(self._events[0], ListDeletingEvent))
        self.assertEqual(self._events[0].mailing_list, mlist)
        self.assertTrue(isinstance(self._events[1], ListDeletedEvent))
        self.assertEqual(self._events[1].fqdn_listname, 'another@example.com')



class TestListLifecycleEvents(unittest.TestCase):
    layer = ConfigLayer

    def setUp(self):
        self._ant = create_list('ant@example.com')
        self._bee = create_list('bee@example.com')
        self._usermanager = getUtility(IUserManager)

    def test_members_are_deleted_when_mailing_list_is_deleted(self):
        # When a mailing list with members is deleted, all the Member records
        # are also deleted.
        anne = self._usermanager.create_address('anne@example.com')
        bart = self._usermanager.create_address('bart@example.com')
        anne_ant = self._ant.subscribe(anne)
        anne_bee = self._bee.subscribe(anne)
        bart_ant = self._ant.subscribe(bart)
        anne_ant_id = anne_ant.member_id
        anne_bee_id = anne_bee.member_id
        bart_ant_id = bart_ant.member_id
        getUtility(IListManager).delete(self._ant)
        service = getUtility(ISubscriptionService)
        # We deleted the ant@example.com mailing list.  Anne's and Bart's
        # membership in this list should now be removed, but Anne's membership
        # in bee@example.com should still exist.
        self.assertEqual(service.get_member(anne_ant_id), None)
        self.assertEqual(service.get_member(bart_ant_id), None)
        self.assertEqual(service.get_member(anne_bee_id), anne_bee)

    def test_requests_are_deleted_when_mailing_list_is_deleted(self):
        # When a mailing list is deleted, its requests database is deleted
        # too, e.g. all its message hold requests (but not the messages
        # themselves).
        msg = specialized_message_from_string("""\
From: anne@example.com
To: ant@example.com
Subject: Hold me
Message-ID: <argon>

""")
        request_id = hold_message(self._ant, msg)
        getUtility(IListManager).delete(self._ant)
        # This is a hack.  ListRequests don't access self._mailinglist in
        # their get_request() method.
        requestsdb = IListRequests(self._bee)
        request = requestsdb.get_request(request_id)
        self.assertEqual(request, None)
        saved_message = getUtility(IMessageStore).get_message_by_id('<argon>')
        self.assertEqual(saved_message.as_string(), msg.as_string())



# Copyright (C) 2009-2012 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""Email address validation."""

from __future__ import absolute_import, print_function, unicode_literals

__metaclass__ = type
__all__ = [
    'Validator',
    ]


import re

from zope.interface import implements

from mailman.interfaces.address import (
    IEmailValidator, InvalidEmailAddressError)
from mailman.utilities.email import split_email


# What other characters should be disallowed?
_badchars = re.compile(r'[][()<>|;^,\000-\037\177-\377]')



class Validator:
    """An email address validator."""

    implements(IEmailValidator)

    def is_valid(self, email):
        """See `IEmailValidator`."""
        if not email or ' ' in email:
            return False
        if _badchars.search(email) or email[0] == '-':
            return False
        user, domain_parts = split_email(email)
        # Local, unqualified addresses are not allowed.
        if not domain_parts:
            return False
        if len(domain_parts) < 2:
            return False
        return True

    def validate(self, email):
        """Validate an email address.

        :param address: An email address.
        :type address: string
        :raise InvalidEmailAddressError: when the address is deemed invalid.
        """
        if not self.is_valid(email):
            raise InvalidEmailAddressError(email)



# Copyright (C) 2007-2012 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""Implementations of the IPendable and IPending interfaces."""

from __future__ import absolute_import, unicode_literals

__metaclass__ = type
__all__ = [
    'Pended',
    'Pendings',
    ]


import time
import random
import hashlib
import datetime

from lazr.config import as_timedelta
from storm.locals import DateTime, Int, RawStr, ReferenceSet, Unicode
from zope.interface import implements
from zope.interface.verify import verifyObject

from mailman.config import config
from mailman.database.model import Model
from mailman.interfaces.pending import (
    IPendable, IPended, IPendedKeyValue, IPendings)
from mailman.utilities.modules import call_name



class PendedKeyValue(Model):
    """A pended key/value pair, tied to a token."""

    implements(IPendedKeyValue)

    def __init__(self, key, value):
        self.key = key
        self.value = value

    id = Int(primary=True)
    key = Unicode()
    value = Unicode()
    pended_id = Int()


class Pended(Model):
    """A pended event, tied to a token."""

    implements(IPended)

    def __init__(self, token, expiration_date):
        super(Pended, self).__init__()
        self.token = token
        self.expiration_date = expiration_date

    id = Int(primary=True)
    token = RawStr()
    expiration_date = DateTime()
    key_values = ReferenceSet(id, PendedKeyValue.pended_id)



class UnpendedPendable(dict):
    implements(IPendable)



class Pendings:
    """Implementation of the IPending interface."""

    implements(IPendings)

    def add(self, pendable, lifetime=None):
        verifyObject(IPendable, pendable)
        # Calculate the token and the lifetime.
        if lifetime is None:
            lifetime = as_timedelta(config.mailman.pending_request_life)
        # Calculate a unique token.  Algorithm vetted by the Timbot.  time()
        # has high resolution on Linux, clock() on Windows.  random gives us
        # about 45 bits in Python 2.2, 53 bits on Python 2.3.  The time and
        # clock values basically help obscure the random number generator, as
        # does the hash calculation.  The integral parts of the time values
        # are discarded because they're the most predictable bits.
        for attempts in range(3):
            now = time.time()
            x = random.random() + now % 1.0 + time.clock() % 1.0
            # Use sha1 because it produces shorter strings.
            token = hashlib.sha1(repr(x)).hexdigest()
            # In practice, we'll never get a duplicate, but we'll be anal
            # about checking anyway.
            if config.db.store.find(Pended, token=token).count() == 0:
                break
        else:
            raise AssertionError('Could not find a valid pendings token')
        # Create the record, and then the individual key/value pairs.
        pending = Pended(
            token=token,
            expiration_date=datetime.datetime.now() + lifetime)
        for key, value in pendable.items():
            if isinstance(key, str):
                key = unicode(key, 'utf-8')
            if isinstance(value, str):
                value = unicode(value, 'utf-8')
            elif type(value) is int:
                value = '__builtin__.int\1%s' % value
            elif type(value) is float:
                value = '__builtin__.float\1%s' % value
            elif type(value) is bool:
                value = '__builtin__.bool\1%s' % value
            elif type(value) is list:
                # We expect this to be a list of strings.
                value = ('mailman.model.pending.unpack_list\1' +
                         '\2'.join(value))
            keyval = PendedKeyValue(key=key, value=value)
            pending.key_values.add(keyval)
        config.db.store.add(pending)
        return token

    def confirm(self, token, expunge=True):
        store = config.db.store
        # Token can come in as a unicode, but it's stored in the database as
        # bytes.  They must be ascii.
        pendings = store.find(Pended, token=str(token))
        if pendings.count() == 0:
            return None
        assert pendings.count() == 1, (
            'Unexpected token count: {0}'.format(pendings.count()))
        pending = pendings[0]
        pendable = UnpendedPendable()
        # Find all PendedKeyValue entries that are associated with the pending
        # object's ID.  Watch out for type conversions.
        for keyvalue in store.find(PendedKeyValue,
                                   PendedKeyValue.pended_id == pending.id):
            if keyvalue.value is not None and '\1' in keyvalue.value:
                type_name, value = keyvalue.value.split('\1', 1)
                pendable[keyvalue.key] = call_name(type_name, value)
            else:
                pendable[keyvalue.key] = keyvalue.value
            if expunge:
                store.remove(keyvalue)
        if expunge:
            store.remove(pending)
        return pendable

    def evict(self):
        store = config.db.store
        now = datetime.datetime.now()
        for pending in store.find(Pended):
            if pending.expiration_date < now:
                # Find all PendedKeyValue entries that are associated with the
                # pending object's ID.
                q = store.find(PendedKeyValue,
                               PendedKeyValue.pended_id == pending.id)
                for keyvalue in q:
                    store.remove(keyvalue)
                store.remove(pending)



def unpack_list(value):
    return value.split('\2')



# Copyright (C) 1998-2012 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""Legacy Mailman exceptions.

This module is largely obsolete, though not all exceptions in use have been
migrated to their proper location.  There are still a number of Mailman 2.1
exceptions floating about in here too.

The right place for exceptions is in the interface module for their related
interfaces.
"""


from __future__ import absolute_import, unicode_literals

__metaclass__ = type
__all__ = [
    'AlreadyReceivingDigests',
    'AlreadyReceivingRegularDeliveries',
    'BadPasswordSchemeError',
    'CantDigestError',
    'DiscardMessage',
    'HandlerError',
    'HoldMessage',
    'LostHeldMessage',
    'MailmanError',
    'MailmanException',
    'MemberError',
    'MustDigestError',
    'PasswordError',
    'RejectMessage',
    ]



# Base class for all exceptions raised in Mailman.
class MailmanException(Exception):
    pass



# "New" style membership exceptions (new w/ MM2.1)
class MemberError(MailmanException): pass
class AlreadyReceivingDigests(MemberError): pass
class AlreadyReceivingRegularDeliveries(MemberError): pass
class CantDigestError(MemberError): pass
class MustDigestError(MemberError): pass



# New style class based exceptions.  All the above errors should eventually be
# converted.

class MailmanError(MailmanException):
    """Base class for all Mailman errors."""
    pass



# Exceptions for admin request database
class LostHeldMessage(MailmanError):
    """Held message was lost."""
    pass



def _(s):
    return s


# Exceptions for the Handler subsystem
class HandlerError(MailmanError):
    """Base class for all handler errors."""


class HoldMessage(HandlerError):
    """Base class for all message-being-held short circuits."""

    # funky spelling is necessary to break import loops
    reason = _('For some unknown reason')

    def reason_notice(self):
        return self.reason

    # funky spelling is necessary to break import loops
    rejection = _('Your message was rejected')

    def rejection_notice(self, mlist):
        return self.rejection


class DiscardMessage(HandlerError):
    """The message can be discarded with no further action"""


class RejectMessage(HandlerError):
    """The message will be bounced back to the sender"""



class PasswordError(MailmanError):
    """A password related error."""


class BadPasswordSchemeError(PasswordError):
    """A bad password scheme was given."""

    def __init__(self, scheme_name='unknown'):
        super(BadPasswordSchemeError, self).__init__()
        self.scheme_name = scheme_name

    def __str__(self):
        return 'A bad password scheme was given: %s' % self.scheme_name



# Copyright (C) 2011-2012 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""RFC 2369 List-* and related headers."""

from __future__ import absolute_import, unicode_literals

__metaclass__ = type
__all__ = [
    'RFC2369',
    ]


from email.utils import formataddr
from zope.interface import implements

from mailman.config import config
from mailman.core.i18n import _
from mailman.handlers.cook_headers import uheader
from mailman.interfaces.handler import IHandler


CONTINUATION = ',\n\t'



def process(mlist, msg, msgdata):
    """Add the RFC 2369 List-* and related headers."""
    # Some people really hate the List-* headers.  It seems that the free
    # version of Eudora (possibly on for some platforms) does not hide these
    # headers by default, pissing off their users.  Too bad.  Fix the MUAs.
    if not mlist.include_rfc2369_headers:
        return
    list_id = '{0.list_name}.{0.mail_host}'.format(mlist)
    if mlist.description:
        # Don't wrap the header since here we just want to get it properly RFC
        # 2047 encoded.
        i18ndesc = uheader(mlist, mlist.description, 'List-Id', maxlinelen=998)
        listid_h = formataddr((str(i18ndesc), list_id))
    else:
        # Without a description, we need to ensure the MUST brackets.
        listid_h = '<{0}>'.format(list_id)
    # No other agent should add a List-ID header except Mailman.
    del msg['list-id']
    msg['List-Id'] = listid_h
    # For internally crafted messages, we also add a (nonstandard),
    # "X-List-Administrivia: yes" header.  For all others (i.e. those coming
    # from list posts), we add a bunch of other RFC 2369 headers.
    requestaddr = mlist.request_address
    subfieldfmt = '<{0}>, <mailto:{1}>'
    listinfo = mlist.script_url('listinfo')
    headers = {}
    # XXX reduced_list_headers used to suppress List-Help, List-Subject, and
    # List-Unsubscribe from UserNotification.  That doesn't seem to make sense
    # any more, so always add those three headers (others will still be
    # suppressed).
    headers.update({
        'List-Help'       : '<mailto:{0}?subject=help>'.format(requestaddr),
        'List-Unsubscribe': subfieldfmt.format(listinfo, mlist.leave_address),
        'List-Subscribe'  : subfieldfmt.format(listinfo, mlist.join_address),
        })
    if not msgdata.get('reduced_list_headers'):
        # List-Post: is controlled by a separate attribute
        if mlist.include_list_post_header:
            headers['List-Post'] = '<mailto:{0}>'.format(mlist.posting_address)
        # Add RFC 2369 and 5064 archiving headers, if archiving is enabled.
        if mlist.archive:
            for archiver in config.archivers:
                headers['List-Archive'] = '<{0}>'.format(
                    archiver.list_url(mlist))
                permalink = archiver.permalink(mlist, msg)
                if permalink is not None:
                    headers['Archived-At'] = permalink
    # XXX RFC 2369 also defines a List-Owner header which we are not currently
    # supporting, but should.
    for h, v in headers.items():
        # First we delete any pre-existing headers because the RFC permits
        # only one copy of each, and we want to be sure it's ours.
        del msg[h]
        # Wrap these lines if they are too long.  78 character width probably
        # shouldn't be hardcoded, but is at least text-MUA friendly.  The
        # adding of 2 is for the colon-space separator.
        if len(h) + 2 + len(v) > 78:
            v = CONTINUATION.join(v.split(', '))
        msg[h] = v



class RFC2369:
    """Add the RFC 2369 List-* headers."""

    implements(IHandler)

    name = 'rfc-2369'
    description = _('Add the RFC 2369 List-* headers.')

    def process(self, mlist, msg, msgdata):
        """See `IHandler`."""
        process(mlist, msg, msgdata)



# Copyright (C) 1998-2012 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""Add the message to the list's current digest."""

from __future__ import absolute_import, unicode_literals

__metaclass__ = type
__all__ = [
    'ToDigest',
    ]


import os
import datetime

from zope.interface import implements

from mailman.config import config
from mailman.core.i18n import _
from mailman.email.message import Message
from mailman.interfaces.digests import DigestFrequency
from mailman.interfaces.handler import IHandler
from mailman.utilities.mailbox import Mailbox



class ToDigest:
    """Add the message to the digest, possibly sending it."""

    implements(IHandler)

    name = 'to-digest'
    description = _('Add the message to the digest, possibly sending it.')

    def process(self, mlist, msg, msgdata):
        """See `IHandler`."""
        # Short circuit for non-digestable messages.
        if not mlist.digestable or msgdata.get('isdigest'):
            return
        # Open the mailbox that will be used to collect the current digest.
        mailbox_path = os.path.join(mlist.data_path, 'digest.mmdf')
        # Lock the mailbox and append the message.
        with Mailbox(mailbox_path, create=True) as mbox:
            mbox.add(msg)
        # Calculate the current size of the mailbox file.  This will not tell
        # us exactly how big the resulting MIME and rfc1153 digest will
        # actually be, but it's the most easily available metric to decide
        # whether the size threshold has been reached.
        size = os.path.getsize(mailbox_path)
        if size >= mlist.digest_size_threshold * 1024.0:
            # The digest is ready to send.  Because we don't want to hold up
            # this process with crafting the digest, we're going to move the
            # digest file to a safe place, then craft a fake message for the
            # DigestRunner as a trigger for it to build and send the digest.
            mailbox_dest = os.path.join(
                mlist.data_path,
                'digest.{0.volume}.{0.next_digest_number}.mmdf'.format(mlist))
            volume = mlist.volume
            digest_number = mlist.next_digest_number
            bump_digest_number_and_volume(mlist)
            os.rename(mailbox_path, mailbox_dest)
            config.switchboards['digest'].enqueue(
                Message(),
                listname=mlist.fqdn_listname,
                digest_path=mailbox_dest,
                volume=volume,
                digest_number=digest_number)



def bump_digest_number_and_volume(mlist):
    """Bump the digest number and volume."""
    now = datetime.datetime.now()
    if mlist.digest_last_sent_at is None:
        # There has been no previous digest.
        bump = False
    elif mlist.digest_volume_frequency == DigestFrequency.yearly:
        bump = (now.year > mlist.digest_last_sent_at.year)
    elif mlist.digest_volume_frequency == DigestFrequency.monthly:
        # Monthly.
        this_month = now.year * 100 + now.month
        digest_month = (mlist.digest_last_sent_at.year * 100 +
                        mlist.digest_last_sent_at.month)
        bump = (this_month > digest_month)
    elif mlist.digest_volume_frequency == DigestFrequency.quarterly:
        # Quarterly.
        this_quarter = now.year * 100 + (now.month - 1) // 4
        digest_quarter = (mlist.digest_last_sent_at.year * 100 +
                          (mlist.digest_last_sent_at.month - 1) // 4)
        bump = (this_quarter > digest_quarter)
    elif mlist.digest_volume_frequency == DigestFrequency.weekly:
        this_week = now.year * 100 + now.isocalendar()[1]
        digest_week = (mlist.digest_last_sent_at.year * 100 +
                       mlist.digest_last_sent_at.isocalendar()[1])
        bump = (this_week > digest_week)
    elif mlist.digest_volume_frequency == DigestFrequency.daily:
        bump = (now.toordinal() > mlist.digest_last_sent_at.toordinal())
    else:
        raise AssertionError(
            'Bad DigestFrequency: {0}'.format(
                mlist.digest_volume_frequency))
    if bump:
        mlist.volume += 1
        mlist.next_digest_number = 1
    else:
        # Just bump the digest number.
        mlist.next_digest_number += 1
    mlist.digest_last_sent_at = now



# Copyright (C) 2009-2012 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""MTA connections."""

from __future__ import absolute_import, unicode_literals

__metaclass__ = type
__all__ = [
    'Connection',
    ]


import logging
import smtplib

from lazr.config import as_boolean
from mailman.config import config


log = logging.getLogger('mailman.smtp')



class Connection:
    """Manage a connection to the SMTP server."""
    def __init__(self, host, port, sessions_per_connection,
                 smtp_user=None, smtp_pass=None):
        """Create a connection manager.

        :param host: The host name of the SMTP server to connect to.
        :type host: string
        :param port: The port number of the SMTP server to connect to.
        :type port: integer
        :param sessions_per_connection: The number of SMTP sessions per
            connection to the SMTP server.  After this number of sessions
            has been reached, the connection is closed and a new one is
            opened.  Set to zero for an unlimited number of sessions per
            connection (i.e. your MTA has no limit).
        :type sessions_per_connection: integer
        :param smtp_user: Optional SMTP authentication user name.  If given,
            `smtp_pass` must also be given.
        :type smtp_user: str
        :param smtp_pass: Optional SMTP authentication password.  If given,
            `smtp_user` must also be given.
        """
        self._host = host
        self._port = port
        self._sessions_per_connection = sessions_per_connection
        self._username = smtp_user
        self._password = smtp_pass
        self._session_count = None
        self._connection = None

    def _connect(self):
        """Open a new connection."""
        self._connection = smtplib.SMTP()
        log.debug('Connecting to %s:%s', self._host, self._port)
        self._connection.connect(self._host, self._port)
        if self._username is not None and self._password is not None:
            log.debug('Logging in')
            self._connection.login(self._username, self._password)
        self._session_count = self._sessions_per_connection

    def sendmail(self, envsender, recipients, msgtext):
        """Mimic `smtplib.SMTP.sendmail`."""
        if as_boolean(config.devmode.enabled):
            # Force the recipients to the specified address, but still deliver
            # to the same number of recipients.
            recipients = [config.devmode.recipient] * len(recipients)
        if self._connection is None:
            self._connect()
        try:
            log.debug('envsender: %s, recipients: %s, size(msgtext): %s',
                      envsender, recipients, len(msgtext))
            results = self._connection.sendmail(envsender, recipients, msgtext)
        except smtplib.SMTPException:
            # For safety, close this connection.  The next send attempt will
            # automatically re-open it.  Pass the exception on up.
            self.quit()
            raise
        # This session has been successfully completed.
        self._session_count -= 1
        # By testing exactly for equality to 0, we automatically handle the
        # case for SMTP_MAX_SESSIONS_PER_CONNECTION <= 0 meaning never close
        # the connection.  We won't worry about wraparound <wink>.
        if self._session_count == 0:
            self.quit()
        return results

    def quit(self):
        """Mimic `smtplib.SMTP.quit`."""
        if self._connection is None:
            return
        try:
            self._connection.quit()
        except smtplib.SMTPException:
            pass
        self._connection = None



# Copyright (C) 2010-2012 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""Importer routines."""

from __future__ import absolute_import, unicode_literals

__metaclass__ = type
__all__ = [
    'import_config_pck',
    ]


import sys
import datetime

from mailman.interfaces.autorespond import ResponseAction
from mailman.interfaces.digests import DigestFrequency
from mailman.interfaces.mailinglist import Personalization, ReplyToMunging
from mailman.interfaces.nntp import NewsModeration



def seconds_to_delta(value):
    return datetime.timedelta(seconds=value)


# Attributes in Mailman 2 which have a different type in Mailman 3.
TYPES = dict(
    autorespond_owner=ResponseAction,
    autorespond_postings=ResponseAction,
    autorespond_requests=ResponseAction,
    bounce_info_stale_after=seconds_to_delta,
    bounce_you_are_disabled_warnings_interval=seconds_to_delta,
    digest_volume_frequency=DigestFrequency,
    news_moderation=NewsModeration,
    personalize=Personalization,
    reply_goes_to_list=ReplyToMunging,
    )


# Attribute names in Mailman 2 which are renamed in Mailman 3.
NAME_MAPPINGS = dict(
    host_name='mail_host',
    real_name='display_name',
    )



def import_config_pck(mlist, config_dict):
    """Apply a config.pck configuration dictionary to a mailing list.

    :param mlist: The mailing list.
    :type mlist: IMailingList
    :param config_dict: The Mailman 2.1 configuration dictionary.
    :type config_dict: dict
    """
    for key, value in config_dict.items():
        # Some attributes from Mailman 2 were renamed in Mailman 3.
        key = NAME_MAPPINGS.get(key, key)
        # Handle the simple case where the key is an attribute of the
        # IMailingList and the types are the same (modulo 8-bit/unicode
        # strings).
        if hasattr(mlist, key):
            if isinstance(value, str):
                value = unicode(value, 'ascii')
            # Some types require conversion.
            converter = TYPES.get(key)
            if converter is not None:
                value = converter(value)
            try:
                setattr(mlist, key, value)
            except TypeError:
                print >> sys.stderr, 'Type conversion error:', key
                raise



# Copyright (C) 2001-2012 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""Creation/deletion hooks for the Postfix MTA."""

from __future__ import absolute_import, unicode_literals

__metaclass__ = type
__all__ = [
    'LMTP',
    ]


import os
import logging
import datetime

from flufl.lock import Lock
from operator import attrgetter
from zope.component import getUtility
from zope.interface import implements

from mailman.config import config
from mailman.interfaces.listmanager import IListManager
from mailman.interfaces.mta import (
    IMailTransportAgentAliases, IMailTransportAgentLifecycle)


log = logging.getLogger('mailman.error')
ALIASTMPL = '{0:{2}}lmtp:[{1.mta.lmtp_host}]:{1.mta.lmtp_port}'



class _FakeList:
    """Duck-typed list for the `IMailTransportAgentAliases` interface."""

    def __init__(self, list_name, mail_host):
        self.list_name = list_name
        self.mail_host = mail_host
        self.posting_address = '{0}@{1}'.format(list_name, mail_host)



class LMTP:
    """Connect Mailman to Postfix via LMTP."""

    implements(IMailTransportAgentLifecycle)

    def create(self, mlist):
        """See `IMailTransportAgentLifecycle`."""
        # We can ignore the mlist argument because for LMTP delivery, we just
        # generate the entire file every time.
        self.regenerate()

    delete = create

    def regenerate(self, output=None):
        """See `IMailTransportAgentLifecycle`.

        The format for Postfix's LMTP transport map is defined here:
        http://www.postfix.org/transport.5.html
        """
        # Acquire a lock file to prevent other processes from racing us here.
        lock_file = os.path.join(config.LOCK_DIR, 'mta')
        with Lock(lock_file):
            # If output is a filename, open up a backing file and write the
            # output there, then do the atomic rename dance.  First though, if
            # it's None, we use a calculated path.
            if output is None:
                path = os.path.join(config.DATA_DIR, 'postfix_lmtp')
                path_new = path + '.new'
            elif isinstance(output, basestring):
                path = output
                path_new = output + '.new'
            else:
                path = path_new = None
            if path_new is None:
                self._do_write_file(output)
                # There's nothing to rename, and we can't generate the .db
                # file, so we're done.
                return
            # Write the file.
            with open(path_new, 'w') as fp:
                self._do_write_file(fp)
            # Atomically rename to the intended path.
            os.rename(path + '.new', path)
            # Now that the new file is in place, we must tell Postfix to
            # generate a new .db file.
            command = config.mta.postfix_map_cmd + ' ' + path
            status = (os.system(command) >> 8) & 0xff
            if status:
                msg = 'command failure: %s, %s, %s'
                errstr = os.strerror(status)
                log.error(msg, command, status, errstr)
                raise RuntimeError(msg % (command, status, errstr))

    def _do_write_file(self, fp):
        """Do the actual file writes for list creation."""
        # Sort all existing mailing list names first by domain, then by local
        # part.  For postfix we need a dummy entry for the domain.
        list_manager = getUtility(IListManager)
        by_domain = {}
        for list_name, mail_host in list_manager.name_components:
            mlist = _FakeList(list_name, mail_host)
            by_domain.setdefault(mlist.mail_host, []).append(mlist)
        print >> fp, """\
# AUTOMATICALLY GENERATED BY MAILMAN ON {0}
#
# This file is generated by Mailman, and is kept in sync with the binary hash
# file.  YOU SHOULD NOT MANUALLY EDIT THIS FILE unless you know what you're
# doing, and can keep the two files properly in sync.  If you screw it up,
# you're on your own.
""".format(datetime.datetime.now().replace(microsecond=0))
        sort_key = attrgetter('list_name')
        for domain in sorted(by_domain):
            print >> fp, """\
# Aliases which are visible only in the @{0} domain.""".format(domain)
            for mlist in sorted(by_domain[domain], key=sort_key):
                utility = getUtility(IMailTransportAgentAliases)
                aliases = list(utility.aliases(mlist))
                width = max(len(alias) for alias in aliases) + 3
                print >> fp, ALIASTMPL.format(aliases.pop(0), config, width)
                for alias in aliases:
                    print >> fp, ALIASTMPL.format(alias, config, width)
                print >> fp



# Copyright (C) 2009-2012 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""Module stuff."""

from __future__ import absolute_import, unicode_literals

__metaclass__ = type
__all__ = [
    'SubscriptionService',
    'handle_ListDeletedEvent',
    ]


from flufl.password import generate
from operator import attrgetter
from storm.expr import And, Or
from uuid import UUID
from zope.component import getUtility
from zope.interface import implements

from mailman.app.membership import add_member, delete_member
from mailman.config import config
from mailman.core.constants import system_preferences
from mailman.interfaces.address import IEmailValidator
from mailman.interfaces.listmanager import (
    IListManager, ListDeletedEvent, NoSuchListError)
from mailman.interfaces.member import DeliveryMode, MemberRole
from mailman.interfaces.subscriptions import (
    ISubscriptionService, MissingUserError)
from mailman.interfaces.usermanager import IUserManager
from mailman.model.member import Member



def _membership_sort_key(member):
    """Sort function for get_members().

    The members are sorted first by fully-qualified mailing list name,
    then by subscribed email address, then by role.
    """
    # member.mailing_list is already the fqdn_listname, not the IMailingList
    # object.
    return (member.mailing_list,
            member.address.email,
            int(member.role))



class SubscriptionService:
    """Subscription services for the REST API."""

    implements(ISubscriptionService)

    __name__ = 'members'

    def get_members(self):
        """See `ISubscriptionService`."""
        # {fqdn_listname -> {role -> [members]}}
        by_list = {}
        user_manager = getUtility(IUserManager)
        for member in user_manager.members:
            by_role = by_list.setdefault(member.mailing_list, {})
            members = by_role.setdefault(member.role.name, [])
            members.append(member)
        # Flatten into single list sorted as per the interface.
        all_members = []
        address_of_member = attrgetter('address.email')
        for fqdn_listname in sorted(by_list):
            by_role = by_list[fqdn_listname]
            all_members.extend(
                sorted(by_role.get('owner', []), key=address_of_member))
            all_members.extend(
                sorted(by_role.get('moderator', []), key=address_of_member))
            all_members.extend(
                sorted(by_role.get('member', []), key=address_of_member))
        return all_members

    def get_member(self, member_id):
        """See `ISubscriptionService`."""
        members = config.db.store.find(
            Member,
            Member._member_id == member_id)
        if members.count() == 0:
            return None
        else:
            assert members.count() == 1, 'Too many matching members'
            return members[0]

    def find_members(self, subscriber=None, fqdn_listname=None, role=None):
        """See `ISubscriptionService`."""
        # If `subscriber` is a user id, then we'll search for all addresses
        # which are controlled by the user, otherwise we'll just search for
        # the given address.
        user_manager = getUtility(IUserManager)
        if subscriber is None and fqdn_listname is None and role is None:
            return []
        # Querying for the subscriber is the most complicated part, because
        # the parameter can either be an email address or a user id.
        query = []
        if subscriber is not None:
            if isinstance(subscriber, basestring):
                # subscriber is an email address.
                address = user_manager.get_address(subscriber)
                user = user_manager.get_user(subscriber)
                # This probably could be made more efficient.
                if address is None or user is None:
                    return []
                query.append(Or(Member.address_id == address.id,
                                Member.user_id == user.id))
            else:
                # subscriber is a user id.
                user = user_manager.get_user_by_id(subscriber)
                address_ids = list(address.id for address in user.addresses
                                   if address.id is not None)
                if len(address_ids) == 0 or user is None:
                    return []
                query.append(Or(Member.user_id == user.id,
                                Member.address_id.is_in(address_ids)))
        # Calculate the rest of the query expression, which will get And'd
        # with the Or clause above (if there is one).
        if fqdn_listname is not None:
            query.append(Member.mailing_list == fqdn_listname)
        if role is not None:
            query.append(Member.role == role)
        results = config.db.store.find(Member, And(*query))
        return sorted(results, key=_membership_sort_key)

    def __iter__(self):
        for member in self.get_members():
            yield member

    def join(self, fqdn_listname, subscriber,
             display_name=None,
             delivery_mode=DeliveryMode.regular,
             role=MemberRole.member):
        """See `ISubscriptionService`."""
        mlist = getUtility(IListManager).get(fqdn_listname)
        if mlist is None:
            raise NoSuchListError(fqdn_listname)
        # Is the subscriber an email address or user id?
        if isinstance(subscriber, basestring):
            # It's an email address, so we'll want a real name.  Make sure
            # it's a valid email address, and let InvalidEmailAddressError
            # propagate up.
            getUtility(IEmailValidator).validate(subscriber)
            if display_name is None:
                display_name, at, domain = subscriber.partition('@')
            # Because we want to keep the REST API simple, there is no
            # password or language given to us.  We'll use the system's
            # default language for the user's default language.  We'll set the
            # password to a system default.  This will have to get reset since
            # it can't be retrieved.  Note that none of these are used unless
            # the address is completely new to us.
            password = generate(int(config.passwords.password_length))
            return add_member(mlist, subscriber, display_name, password,
                              delivery_mode,
                              system_preferences.preferred_language, role)
        else:
            # We have to assume it's a UUID.
            assert isinstance(subscriber, UUID), 'Not a UUID'
            user = getUtility(IUserManager).get_user_by_id(subscriber)
            if user is None:
                raise MissingUserError(subscriber)
            return mlist.subscribe(user, role)

    def leave(self, fqdn_listname, email):
        """See `ISubscriptionService`."""
        mlist = getUtility(IListManager).get(fqdn_listname)
        if mlist is None:
            raise NoSuchListError(fqdn_listname)
        # XXX for now, no notification or user acknowledgment.
        delete_member(mlist, email, False, False)



def handle_ListDeletedEvent(event):
    """Delete a mailing list's members when the list is deleted."""

    if not isinstance(event, ListDeletedEvent):
        return
    # Find all the members still associated with the mailing list.
    members = getUtility(ISubscriptionService).find_members(
        fqdn_listname=event.fqdn_listname)
    for member in members:
        member.unsubscribe()



# Copyright (C) 2006-2012 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""Base class for all database classes."""

from __future__ import absolute_import, unicode_literals

__metaclass__ = type
__all__ = [
    'Model',
    ]


import sys

from operator import attrgetter

from storm.properties import PropertyPublisherMeta



class ModelMeta(PropertyPublisherMeta):
    """Do more magic on table classes."""

    _class_registry = set()

    def __init__(self, name, bases, dict):
        # Before we let the base class do it's thing, force an __storm_table__
        # property to enforce our table naming convention.
        self.__storm_table__ = name.lower()
        super(ModelMeta, self).__init__(name, bases, dict)
        # Register the model class so that it can be more easily cleared.
        # This is required by the test framework.
        if name == 'Model':
            return
        ModelMeta._class_registry.add(self)

    @staticmethod
    def _reset(store):
        from mailman.config import config
        from mailman.model.version import Version
        config.db._pre_reset(store)
        # Give each schema migration a chance to do its pre-reset.  See below
        # for calling its post reset too.
        versions = sorted(version.version for version in
                          store.find(Version, component='schema'))
        migrations = {}
        for version in versions:
            # We have to give the migrations module that loaded this version a
            # chance to do both pre- and post-reset operations.  The following
            # find the actual the module path for the migration.  See
            # StormBaseDatabase.load_schema().
            migration = store.find(Version, component=version).one()
            if migration is None:
                continue
            migrations[version] = module_path = migration.version
            module = sys.modules[module_path]
            pre_reset = getattr(module, 'pre_reset', None)
            if pre_reset is not None:
                pre_reset(store)
        # Make sure this is deterministic, by sorting on the storm table name.
        classes = sorted(ModelMeta._class_registry,
                         key=attrgetter('__storm_table__'))
        for model_class in classes:
            store.find(model_class).remove()
        # Now give each migration a chance to do post-reset operations.
        for version in versions:
            module = sys.modules[migrations[version]]
            post_reset = getattr(module, 'post_reset', None)
            if post_reset is not None:
                post_reset(store)
        config.db._post_reset(store)



class Model:
    """Like Storm's `Storm` subclass, but with a bit extra."""
    __metaclass__ = ModelMeta



# Copyright (C) 2007-2012 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""Interfaces for the pending database.

The pending database contains events that must be confirmed by the user.  It
maps these events to a unique hash that can be used as a token for end user
confirmation.
"""

from __future__ import absolute_import, unicode_literals

__metaclass__ = type
__all__ = [
    'IPendable',
    'IPended',
    'IPendedKeyValue',
    'IPendings',
    ]


from zope.interface import Interface, Attribute



class IPendable(Interface):
    """A pendable object."""

    def keys():
        """The keys of the pending event data, all of which are strings."""

    def values():
        """The values of the pending event data, all of which are strings."""

    def items():
        """The key/value pairs of the pending event data.

        Both the keys and values must be strings.
        """



class IPended(Interface):
    """A pended event, tied to a token."""

    token = Attribute("""The pended token.""")

    expiration_date = Attribute("""The expiration date of the pended event.""")



class IPendedKeyValue(Interface):
    """A pended key/value pair."""

    key = Attribute("""The pended key.""")

    value = Attribute("""The pended value.""")



class IPendings(Interface):
    """Interface to pending database."""

    def add(pendable, lifetime=None):
        """Create a new entry in the pending database, returning a token.

        :param pendable: The IPendable instance to add.
        :param lifetime: The amount of time, as a `datetime.timedelta` that
            the pended item should remain in the database.  When None is
            given, a system default maximum lifetime is used.
        :return: A token string for inclusion in urls and email confirmations.
        """

    def confirm(token, expunge=True):
        """Return the IPendable matching the token.

        :param token: The token string for the IPendable given by the `.add()`
            method.
        :param expunge: A flag indicating whether the pendable record should
            also be removed from the database or not.
        :return: The matching IPendable or None if no match was found.
        """

    def evict():
        """Remove all pended items whose lifetime has expired."""



# Copyright (C) 2012 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""The standard -owner posting chain."""

from __future__ import absolute_import, print_function, unicode_literals

__metaclass__ = type
__all__ = [
    'BuiltInOwnerChain',
    ]


import logging

from zope.event import notify

from mailman.chains.base import ChainNotification, TerminalChainBase
from mailman.config import config
from mailman.core.i18n import _


log = logging.getLogger('mailman.vette')



class OwnerNotification(ChainNotification):
    """An event signaling that a message is accepted to the -owner address."""



class BuiltInOwnerChain(TerminalChainBase):
    """Default built-in -owner address chain."""

    name = 'default-owner-chain'
    description = _('The built-in -owner posting chain.')

    def _process(self, mlist, msg, msgdata):
        # At least for now, everything posted to -owners goes through.
        config.switchboards['pipeline'].enqueue(msg, msgdata)
        log.info('OWNER: %s', msg.get('message-id', 'n/a'))
        notify(OwnerNotification(mlist, msg, msgdata, self))



# Copyright (C) 2006-2012 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""Provide an interactive prompt, mimicking the Python interpreter."""

from __future__ import unicode_literals

__metaclass__ = type
__all__ = [
    'interact',
    ]


import os
import sys
import code

DEFAULT_BANNER = ''



def interact(upframe=True, banner=DEFAULT_BANNER, overrides=None):
    """Start an interactive interpreter prompt.

    :param upframe: Whether or not to populate the interpreter's globals with
        the locals from the frame that called this function.
    :type upfframe: bool
    :param banner: The banner to print before the interpreter starts.
    :type banner: string
    :param overrides: Additional interpreter globals to add.
    :type overrides: dict
    """
    # The interactive prompt's namespace.
    namespace = dict()
    # Populate the console's with the locals of the frame that called this
    # function (i.e. one up from here).
    if upframe:
        frame = sys._getframe(1)
        namespace.update(frame.f_globals)
        namespace.update(frame.f_locals)
    if overrides is not None:
        namespace.update(overrides)
    interp = code.InteractiveConsole(namespace)
    # Try to import the readline module, but don't worry if it's unavailable.
    try:
        import readline
    except ImportError:
        pass
    # Mimic the real interactive interpreter's loading of any $PYTHONSTARTUP
    # file.  Note that if the startup file is not prepared to be exec'd more
    # than once, this could cause a problem.
    startup = os.environ.get('PYTHONSTARTUP')
    if startup:
        try:
            execfile(startup, namespace)
        except:
            pass
    # We don't want the funky console object in parentheses in the banner.
    if banner == DEFAULT_BANNER:
        banner = '''\
Python %s on %s
Type "help", "copyright", "credits" or "license" for more information.''' % (
            sys.version, sys.platform)
    elif not banner:
        banner = None
    interp.interact(banner)
    # When an exception occurs in the InteractiveConsole, the various
    # sys.exc_* attributes get set so that error handling works the same way
    # there as it does in the built-in interpreter.  Be anal about clearing
    # any exception information before we're done.
    sys.exc_clear()
    sys.last_type = sys.last_value = sys.last_traceback = None



# Copyright (C) 2007-2012 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""Model for members."""

from __future__ import absolute_import, unicode_literals

__metaclass__ = type
__all__ = [
    'Member',
    ]

from storm.locals import Int, Reference, Unicode
from storm.properties import UUID
from zope.component import getUtility
from zope.interface import implements

from mailman.config import config
from mailman.core.constants import system_preferences
from mailman.database.model import Model
from mailman.database.types import Enum
from mailman.interfaces.action import Action
from mailman.interfaces.address import IAddress
from mailman.interfaces.listmanager import IListManager
from mailman.interfaces.member import IMember, MemberRole, MembershipError
from mailman.interfaces.user import IUser, UnverifiedAddressError
from mailman.interfaces.usermanager import IUserManager
from mailman.utilities.uid import UniqueIDFactory


uid_factory = UniqueIDFactory(context='members')



class Member(Model):
    implements(IMember)

    id = Int(primary=True)
    _member_id = UUID()
    role = Enum(MemberRole)
    mailing_list = Unicode()
    moderation_action = Enum(Action)

    address_id = Int()
    _address = Reference(address_id, 'Address.id')
    preferences_id = Int()
    preferences = Reference(preferences_id, 'Preferences.id')
    user_id = Int()
    _user = Reference(user_id, 'User.id')

    def __init__(self, role, mailing_list, subscriber):
        self._member_id = uid_factory.new_uid()
        self.role = role
        self.mailing_list = mailing_list
        if IAddress.providedBy(subscriber):
            self._address = subscriber
            # Look this up dynamically.
            self._user = None
        elif IUser.providedBy(subscriber):
            self._user = subscriber
            # Look this up dynamically.
            self._address = None
        else:
            raise ValueError('subscriber must be a user or address')
        if role in (MemberRole.owner, MemberRole.moderator):
            self.moderation_action = Action.accept
        elif role is MemberRole.member:
            self.moderation_action = getUtility(IListManager).get(
                mailing_list).default_member_action
        else:
            assert role is MemberRole.nonmember, (
                'Invalid MemberRole: {0}'.format(role))
            self.moderation_action = getUtility(IListManager).get(
                mailing_list).default_nonmember_action

    def __repr__(self):
        return '<Member: {0} on {1} as {2}>'.format(
            self.address, self.mailing_list, self.role)

    @property
    def member_id(self):
        """See `IMember`."""
        return self._member_id

    @property
    def address(self):
        """See `IMember`."""
        return (self._user.preferred_address
                if self._address is None
                else self._address)

    @address.setter
    def address(self, new_address):
        """See `IMember`."""
        if self._address is None:
            # XXX Either we need a better exception here, or we should allow
            # changing a subscription from preferred address to explicit
            # address (and vice versa via del'ing the .address attribute.
            raise MembershipError('Membership is via preferred address')
        if new_address.verified_on is None:
            # A member cannot change their subscription address to an
            # unverified address.
            raise UnverifiedAddressError(new_address)
        user = getUtility(IUserManager).get_user(new_address.email)
        if user is None or user != self.user:
            raise MembershipError('Address is not controlled by user')
        self._address = new_address

    @property
    def user(self):
        """See `IMember`."""
        return (self._user
                if self._address is None
                else getUtility(IUserManager).get_user(self._address.email))

    def _lookup(self, preference):
        pref = getattr(self.preferences, preference)
        if pref is not None:
            return pref
        pref = getattr(self.address.preferences, preference)
        if pref is not None:
            return pref
        if self.address.user:
            pref = getattr(self.address.user.preferences, preference)
            if pref is not None:
                return pref
        return getattr(system_preferences, preference)

    @property
    def acknowledge_posts(self):
        """See `IMember`."""
        return self._lookup('acknowledge_posts')

    @property
    def preferred_language(self):
        """See `IMember`."""
        return self._lookup('preferred_language')

    @property
    def receive_list_copy(self):
        """See `IMember`."""
        return self._lookup('receive_list_copy')

    @property
    def receive_own_postings(self):
        """See `IMember`."""
        return self._lookup('receive_own_postings')

    @property
    def delivery_mode(self):
        """See `IMember`."""
        return self._lookup('delivery_mode')

    @property
    def delivery_status(self):
        """See `IMember`."""
        return self._lookup('delivery_status')

    @property
    def options_url(self):
        """See `IMember`."""
        # XXX Um, this is definitely wrong
        return 'http://example.com/' + self.address.email

    def unsubscribe(self):
        """See `IMember`."""
        config.db.store.remove(self.preferences)
        config.db.store.remove(self)



# Copyright (C) 2009-2012 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""Digest runner."""

from __future__ import absolute_import, unicode_literals

__metaclass__ = type
__all__ = [
    'DigestRunner',
    ]


import re
import logging

# cStringIO doesn't support unicode.
from StringIO import StringIO
from contextlib import nested
from copy import deepcopy
from email.header import Header
from email.message import Message
from email.mime.message import MIMEMessage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.utils import formatdate, getaddresses, make_msgid
from urllib2 import URLError

from mailman.config import config
from mailman.core.i18n import _
from mailman.core.runner import Runner
from mailman.handlers.decorate import decorate
from mailman.interfaces.member import DeliveryMode, DeliveryStatus
from mailman.utilities.i18n import make
from mailman.utilities.mailbox import Mailbox
from mailman.utilities.string import oneline, wrap


log = logging.getLogger('mailman.error')



class Digester:
    """Base digester class."""

    def __init__(self, mlist, volume, digest_number):
        self._mlist = mlist
        self._charset = mlist.preferred_language.charset
        # This will be used in the Subject, so use $-strings.
        self._digest_id = _(
            '$mlist.display_name Digest, Vol $volume, Issue $digest_number')
        self._subject = Header(self._digest_id,
                               self._charset,
                               header_name='Subject')
        self._message = self._make_message()
        self._message['From'] = mlist.request_address
        self._message['Subject'] = self._subject
        self._message['To'] = mlist.posting_address
        self._message['Reply-To'] = mlist.posting_address
        self._message['Date'] = formatdate(localtime=True)
        self._message['Message-ID'] = make_msgid()
        # In the rfc1153 digest, the masthead contains the digest boilerplate
        # plus any digest header.  In the MIME digests, the masthead and
        # digest header are separate MIME subobjects.  In either case, it's
        # the first thing in the digest, and we can calculate it now, so go
        # ahead and add it now.
        self._masthead = make('masthead.txt',
                              mailing_list=mlist,
                              display_name=mlist.display_name,
                              got_list_email=mlist.posting_address,
                              got_listinfo_url=mlist.script_url('listinfo'),
                              got_request_email=mlist.request_address,
                              got_owner_email=mlist.owner_address,
                              )
        # Set things up for the table of contents.
        if mlist.digest_header_uri is not None:
            try:
                self._header = decorate(mlist, mlist.digest_header_uri)
            except URLError:
                log.exception(
                    'Digest header decorator URI not found ({0}): {1}'.format(
                        mlist.fqdn_listname, mlist.digest_header_uri))
                self._header = ''
        self._toc = StringIO()
        print >> self._toc, _("Today's Topics:\n")

    def add_to_toc(self, msg, count):
        """Add a message to the table of contents."""
        subject = msg.get('subject', _('(no subject)'))
        subject = oneline(subject, in_unicode=True)
        # Don't include the redundant subject prefix in the toc
        mo = re.match('(re:? *)?({0})'.format(
            re.escape(self._mlist.subject_prefix)),
                      subject, re.IGNORECASE)
        if mo:
            subject = subject[:mo.start(2)] + subject[mo.end(2):]
        # Take only the first author we find.
        username = ''
        addresses = getaddresses(
            [oneline(msg.get('from', ''), in_unicode=True)])
        if addresses:
            username = addresses[0][0]
            if not username:
                username = addresses[0][1]
        if username:
            username = ' ({0})'.format(username)
        lines = wrap('{0:2}. {1}'. format(count, subject), 65).split('\n')
        # See if the user's name can fit on the last line
        if len(lines[-1]) + len(username) > 70:
            lines.append(username)
        else:
            lines[-1] += username
        # Add this subject to the accumulating topics
        first = True
        for line in lines:
            if first:
                print >> self._toc, ' ', line
                first = False
            else:
                print >> self._toc, '     ', line.lstrip()

    def add_message(self, msg, count):
        """Add the message to the digest."""
        # We do not want all the headers of the original message to leak
        # through in the digest messages.
        keepers = {}
        for header in self._keepers:
            keepers[header] = msg.get_all(header, [])
        # Remove all the unkempt <wink> headers.  Use .keys() to allow for
        # destructive iteration...
        for header in msg.keys():
            del msg[header]
        # ... and add them in the designated order.
        for header in self._keepers:
            for value in keepers[header]:
                msg[header] = value
        # Add some useful extra stuff.
        msg['Message'] = unicode(count)




class MIMEDigester(Digester):
    """A MIME digester."""

    def __init__(self, mlist, volume, digest_number):
        super(MIMEDigester, self).__init__(mlist, volume, digest_number)
        masthead = MIMEText(self._masthead.encode(self._charset),
                            _charset=self._charset)
        masthead['Content-Description'] = self._subject
        self._message.attach(masthead)
        # Add the optional digest header.
        if mlist.digest_header_uri is not None:
            header = MIMEText(self._header.encode(self._charset),
                              _charset=self._charset)
            header['Content-Description'] = _('Digest Header')
            self._message.attach(header)
        # Calculate the set of headers we're to keep in the MIME digest.
        self._keepers = set(config.digests.mime_digest_keep_headers.split())

    def _make_message(self):
        return MIMEMultipart('mixed')

    def add_toc(self, count):
        """Add the table of contents."""
        toc_text = self._toc.getvalue()
        try:
            toc_part = MIMEText(toc_text.encode(self._charset),
                                _charset=self._charset)
        except UnicodeError:
            toc_part = MIMEText(toc_text.encode('utf-8'), _charset='utf-8')
        toc_part['Content-Description']= _("Today's Topics ($count messages)")
        self._message.attach(toc_part)

    def add_message(self, msg, count):
        """Add the message to the digest."""
        # Make a copy of the message object, since the RFC 1153 processing
        # scrubs out attachments.
        self._message.attach(MIMEMessage(deepcopy(msg)))

    def finish(self):
        """Finish up the digest, producing the email-ready copy."""
        if self._mlist.digest_footer_uri is not None:
            try:
                footer_text = decorate(
                    self._mlist, self._mlist.digest_footer_uri)
            except URLError:
                log.exception(
                    'Digest footer decorator URI not found ({0}): {1}'.format(
                        self._mlist.fqdn_listname, 
                        self._mlist.digest_footer_uri))
                footer_text = ''
            footer = MIMEText(footer_text.encode(self._charset),
                              _charset=self._charset)
            footer['Content-Description'] = _('Digest Footer')
            self._message.attach(footer)
        # This stuff is outside the normal MIME goo, and it's what the old
        # MIME digester did.  No one seemed to complain, probably because you
        # won't see it in an MUA that can't display the raw message.  We've
        # never got complaints before, but if we do, just wax this.  It's
        # primarily included for (marginally useful) backwards compatibility.
        self._message.postamble = _('End of ') + self._digest_id
        return self._message



class RFC1153Digester(Digester):
    """A digester of the format specified by RFC 1153."""

    def __init__(self, mlist, volume, digest_number):
        super(RFC1153Digester, self).__init__(mlist, volume, digest_number)
        self._separator70 = '-' * 70
        self._separator30 = '-' * 30
        self._text = StringIO()
        print >> self._text, self._masthead
        print >> self._text
        # Add the optional digest header.
        if mlist.digest_header_uri is not None:
            print >> self._text, self._header
            print >> self._text
        # Calculate the set of headers we're to keep in the RFC1153 digest.
        self._keepers = set(config.digests.plain_digest_keep_headers.split())

    def _make_message(self):
        return Message()

    def add_toc(self, count):
        """Add the table of contents."""
        print >> self._text, self._toc.getvalue()
        print >> self._text
        print >> self._text, self._separator70
        print >> self._text

    def add_message(self, msg, count):
        """Add the message to the digest."""
        if count > 1:
            print >> self._text, self._separator30
            print >> self._text
        # Each message section contains a few headers.
        for header in config.digests.plain_digest_keep_headers.split():
            if header in msg:
                value = oneline(msg[header], in_unicode=True)
                value = wrap('{0}: {1}'.format(header, value))
                value = '\n\t'.join(value.split('\n'))
                print >> self._text, value
        print >> self._text
        # Add the payload.  If the decoded payload is empty, this may be a
        # multipart message.  In that case, just stringify it.
        payload = msg.get_payload(decode=True)
        payload = (payload if payload else msg.as_string().split('\n\n', 1)[1])
        try:
            charset = msg.get_content_charset('us-ascii')
            payload = unicode(payload, charset, 'replace')
        except (LookupError, TypeError):
            # Unknown or empty charset.
            payload = unicode(payload, 'us-ascii', 'replace')
        print >> self._text, payload
        if not payload.endswith('\n'):
            print >> self._text

    def finish(self):
        """Finish up the digest, producing the email-ready copy."""
        if self._mlist.digest_footer_uri is not None:
            try:
                footer_text = decorate(
                    self._mlist, self._mlist.digest_footer_uri)
            except URLError:
                log.exception(
                    'Digest footer decorator URI not found ({0}): {1}'.format(
                        self._mlist.fqdn_listname, 
                        self._mlist.digest_footer_uri))
                footer_text = ''                
            # This is not strictly conformant RFC 1153.  The trailer is only
            # supposed to contain two lines, i.e. the "End of ... Digest" line
            # and the row of asterisks.  If this screws up MUAs, the solution
            # is to add the footer as the last message in the RFC 1153 digest.
            # I just hate the way that VM does that and I think it's confusing
            # to users, so don't do it unless there's a clamor.
            print >> self._text, self._separator30
            print >> self._text
            print >> self._text, footer_text
            print >> self._text
        # Add the sign-off.
        sign_off = _('End of ') + self._digest_id
        print >> self._text, sign_off
        print >> self._text, '*' * len(sign_off)
        # If the digest message can't be encoded by the list character set,
        # fall back to utf-8.
        text = self._text.getvalue()
        try:
            self._message.set_payload(text.encode(self._charset),
                                      charset=self._charset)
        except UnicodeError:
            self._message.set_payload(text.encode('utf-8'), charset='utf-8')
        return self._message



class DigestRunner(Runner):
    """The digest runner."""

    def _dispose(self, mlist, msg, msgdata):
        """See `IRunner`."""
        volume = msgdata['volume']
        digest_number = msgdata['digest_number']
        with nested(Mailbox(msgdata['digest_path']),
                    _.using(mlist.preferred_language.code)) as (mailbox,
                                                                language_code):
            # Create the digesters.
            mime_digest = MIMEDigester(mlist, volume, digest_number)
            rfc1153_digest = RFC1153Digester(mlist, volume, digest_number)
            # Cruise through all the messages in the mailbox, first building
            # the table of contents and accumulating Subject: headers and
            # authors.  The question really is whether it's better from a
            # performance and memory footprint to go through the mailbox once
            # and cache the messages in a list, or to cruise through the
            # mailbox twice.  We'll do the latter, but it's a complete guess.
            count = None
            for count, (key, message) in enumerate(mailbox.iteritems(), 1):
                mime_digest.add_to_toc(message, count)
                rfc1153_digest.add_to_toc(message, count)
            assert count is not None, 'No digest messages?'
            # Add the table of contents.
            mime_digest.add_toc(count)
            rfc1153_digest.add_toc(count)
            # Cruise through the set of messages a second time, adding them to
            # the actual digest.
            for count, (key, message) in enumerate(mailbox.iteritems(), 1):
                mime_digest.add_message(message, count)
                rfc1153_digest.add_message(message, count)
            # Finish up the digests.
            mime = mime_digest.finish()
            rfc1153 = rfc1153_digest.finish()
        # Calculate the recipients lists
        mime_recipients = set()
        rfc1153_recipients = set()
        # When someone turns off digest delivery, they will get one last
        # digest to ensure that there will be no gaps in the messages they
        # receive.
        digest_members = set(mlist.digest_members.members)
        for member in digest_members:
            if member.delivery_status <> DeliveryStatus.enabled:
                continue
            # Send the digest to the case-preserved address of the digest
            # members.
            email_address = member.address.original_email
            if member.delivery_mode == DeliveryMode.plaintext_digests:
                rfc1153_recipients.add(email_address)
            elif member.delivery_mode == DeliveryMode.mime_digests:
                mime_recipients.add(email_address)
            else:
                raise AssertionError(
                    'Digest member "{0}" unexpected delivery mode: {1}'.format(
                        email_address, member.delivery_mode))
        # Add also the folks who are receiving one last digest.
        for address, delivery_mode in mlist.last_digest_recipients:
            if delivery_mode == DeliveryMode.plaintext_digests:
                rfc1153_recipients.add(address.original_email)
            elif delivery_mode == DeliveryMode.mime_digests:
                mime_recipients.add(address.original_email)
            else:
                raise AssertionError(
                    'OLD recipient "{0}" unexpected delivery mode: {1}'.format(
                        address, delivery_mode))
        # Send the digests to the virgin queue for final delivery.
        queue = config.switchboards['virgin']
        queue.enqueue(mime,
                      recipients=mime_recipients,
                      listname=mlist.fqdn_listname,
                      isdigest=True)
        queue.enqueue(rfc1153,
                      recipients=rfc1153_recipients,
                      listname=mlist.fqdn_listname,
                      isdigest=True)



# Copyright (C) 2009-2012 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""The 'lists' subcommand."""

from __future__ import absolute_import, unicode_literals

__metaclass__ = type
__all__ = [
    'Create',
    'Lists',
    'Remove',
    ]


from zope.component import getUtility
from zope.interface import implements

from mailman.app.lifecycle import create_list, remove_list
from mailman.config import config
from mailman.core.constants import system_preferences
from mailman.core.i18n import _
from mailman.email.message import UserNotification
from mailman.interfaces.address import (
    IEmailValidator, InvalidEmailAddressError)
from mailman.interfaces.command import ICLISubCommand
from mailman.interfaces.domain import (
    BadDomainSpecificationError, IDomainManager)
from mailman.interfaces.languages import ILanguageManager
from mailman.interfaces.listmanager import IListManager, ListAlreadyExistsError
from mailman.utilities.i18n import make


COMMASPACE = ', '



class Lists:
    """List all mailing lists"""

    implements(ICLISubCommand)

    name = 'lists'

    def add(self, parser, command_parser):
        """See `ICLISubCommand`."""
        command_parser.add_argument(
            '-a', '--advertised',
            default=False, action='store_true',
            help=_(
                'List only those mailing lists that are publicly advertised'))
        command_parser.add_argument(
            '-n', '--names',
            default=False, action='store_true',
            help=_('Show also the list names'))
        command_parser.add_argument(
            '-d', '--descriptions',
            default=False, action='store_true',
            help=_('Show also the list descriptions'))
        command_parser.add_argument(
            '-q', '--quiet',
            default=False, action='store_true',
            help=_('Less verbosity'))
        command_parser.add_argument(
            '--domain',
            action='append', help=_("""\
            List only those mailing lists hosted on the given domain, which
            must be the email host name.  Multiple -d options may be given.
            """))

    def process(self, args):
        """See `ICLISubCommand`."""
        mailing_lists = []
        list_manager = getUtility(IListManager)
        # Gather the matching mailing lists.
        for fqdn_name in sorted(list_manager.names):
            mlist = list_manager.get(fqdn_name)
            if args.advertised and not mlist.advertised:
                continue
            domains = getattr(args, 'domains', None)
            if domains and mlist.mail_host not in domains:
                continue
            mailing_lists.append(mlist)
        # Maybe no mailing lists matched.
        if len(mailing_lists) == 0:
            if not args.quiet:
                print _('No matching mailing lists found')
            return
        count = len(mailing_lists)
        if not args.quiet:
            print _('$count matching mailing lists found:')
        # Calculate the longest identifier.
        longest = 0
        output = []
        for mlist in mailing_lists:
            if args.names:
                identifier = '{0} [{1}]'.format(
                    mlist.fqdn_listname, mlist.display_name)
            else:
                identifier = mlist.fqdn_listname
            longest = max(len(identifier), longest)
            output.append((identifier, mlist.description))
        # Print it out.
        if args.descriptions:
            format_string = '{0:{2}} - {1:{3}}'
        else:
            format_string = '{0:{2}}'
        for identifier, description in output:
            print format_string.format(
                identifier, description, longest, 70 - longest)



class Create:
    """Create a mailing list"""

    implements(ICLISubCommand)

    name = 'create'

    def add(self, parser, command_parser):
        """See `ICLISubCommand`."""
        self.parser = parser
        command_parser.add_argument(
            '--language',
            type=unicode, metavar='CODE', help=_("""\
            Set the list's preferred language to CODE, which must be a
            registered two letter language code."""))
        command_parser.add_argument(
            '-o', '--owner',
            type=unicode, action='append', default=[],
            dest='owners', metavar='OWNER', help=_("""\
            Specify a listowner email address.  If the address is not
            currently registered with Mailman, the address is registered and
            linked to a user.  Mailman will send a confirmation message to the
            address, but it will also send a list creation notice to the
            address.  More than one owner can be specified."""))
        command_parser.add_argument(
            '-n', '--notify',
            default=False, action='store_true',
            help=_("""\
            Notify the list owner by email that their mailing list has been
            created."""))
        command_parser.add_argument(
            '-q', '--quiet',
            default=False, action='store_true',
            help=_('Print less output.'))
        command_parser.add_argument(
            '-d', '--domain',
            default=False, action='store_true',
            help=_("""\
            Register the mailing list's domain if not yet registered."""))
        # Required positional argument.
        command_parser.add_argument(
            'listname', metavar='LISTNAME', nargs=1,
            help=_("""\
            The 'fully qualified list name', i.e. the posting address of the
            mailing list.  It must be a valid email address and the domain
            must be registered with Mailman.  List names are forced to lower
            case."""))

    def process(self, args):
        """See `ICLISubCommand`."""
        language_code = (args.language
                         if args.language is not None
                         else system_preferences.preferred_language.code)
        # Make sure that the selected language code is known.
        if language_code not in getUtility(ILanguageManager).codes:
            self.parser.error(_('Invalid language code: $language_code'))
            return
        assert len(args.listname) == 1, (
            'Unexpected positional arguments: %s' % args.listname)
        # Check to see if the domain exists or not.
        fqdn_listname = args.listname[0]
        listname, at, domain = fqdn_listname.partition('@')
        domain_manager = getUtility(IDomainManager)
        if domain_manager.get(domain) is None and args.domain:
            domain_manager.add(domain)
        # Validate the owner email addresses.  The problem with doing this
        # check in create_list() is that you wouldn't be able to distinguish
        # between an InvalidEmailAddressError for the list name or the
        # owners.  I suppose we could subclass that exception though.
        if args.owners:
            validator = getUtility(IEmailValidator)
            invalid_owners = [owner for owner in args.owners
                              if not validator.is_valid(owner)]
            if invalid_owners:
                invalid = COMMASPACE.join(sorted(invalid_owners))
                self.parser.error(_('Illegal owner addresses: $invalid'))
                return
        try:
            mlist = create_list(fqdn_listname, args.owners)
        except InvalidEmailAddressError:
            self.parser.error(_('Illegal list name: $fqdn_listname'))
            return
        except ListAlreadyExistsError:
            self.parser.error(_('List already exists: $fqdn_listname'))
            return
        except BadDomainSpecificationError, domain:
            self.parser.error(_('Undefined domain: $domain'))
            return
        # Find the language associated with the code, then set the mailing
        # list's preferred language to that.  The changes then must be
        # committed to the database.
        mlist.preferred_language = getUtility(ILanguageManager)[language_code]
        config.db.commit()
        # Do the notification.
        if not args.quiet:
            print _('Created mailing list: $mlist.fqdn_listname')
        if args.notify:
            d = dict(
                listname        = mlist.fqdn_listname,
                admin_url       = mlist.script_url('admin'),
                listinfo_url    = mlist.script_url('listinfo'),
                requestaddr     = mlist.request_address,
                siteowner       = mlist.no_reply_address,
                )
            text = make('newlist.txt', mailing_list=mlist, **d)
            # Set the I18N language to the list's preferred language so the
            # header will match the template language.  Stashing and restoring
            # the old translation context is just (healthy? :) paranoia.
            with _.using(mlist.preferred_language.code):
                msg = UserNotification(
                    args.owners, mlist.no_reply_address,
                    _('Your new mailing list: $fqdn_listname'),
                    text, mlist.preferred_language)
                msg.send(mlist)



class Remove:
    """Remove a mailing list"""

    implements(ICLISubCommand)

    name = 'remove'

    def add(self, parser, command_parser):
        """See `ICLISubCommand`."""
        command_parser.add_argument(
            '-q', '--quiet',
            default=False, action='store_true',
            help=_('Suppress status messages'))
        # Required positional argument.
        command_parser.add_argument(
            'listname', metavar='LISTNAME', nargs=1,
            help=_("""\
            The 'fully qualified list name', i.e. the posting address of the
            mailing list."""))

    def process(self, args):
        """See `ICLISubCommand`."""
        def log(message):
            if not args.quiet:
                print message
        assert len(args.listname) == 1, (
            'Unexpected positional arguments: %s' % args.listname)
        fqdn_listname = args.listname[0]
        mlist = getUtility(IListManager).get(fqdn_listname)
        if mlist is None:
            log(_('No such list: $fqdn_listname'))
            return
        else:
            log(_('Removed list: $fqdn_listname'))
        remove_list(fqdn_listname, mlist)
        config.db.commit()
